#include "gfunct.h"
#include "tablefunct.h"
#include "itablefunct.h"
#include "intools.h"
#include <string.h>

/**
  This constructor initializes attributes to zero values.
*/
gfunct::gfunct()
{
  //  number of expression sets
  neqs = 0L;
  //  constant value
  f = 0.0;
  tfunc=stat;
  eq = NULL;
  var = NULL;
  func = NULL;
  limval = NULL;
  tabf = NULL;
  itabf = NULL;
}

/**
  Destructor deallocates memory used by the expression
*/
gfunct::~gfunct()
{
  for (long i = 0; i < neqs; i++){
    delete [] func[i];
    delete [] var[i];
    delete eq[i];
  }
  delete [] func;
  delete [] eq;
  delete [] var;
  delete tabf;
  delete itabf;
}

/**
  This function reads data from the text file given by the parameter in. The first it reads
  type of function see galias.h - timefunct. Then it depending on the function type it
  reads data about function.
  Parameters :
  @param in - pointer to opened text file from which data will be read.
  Return :
  @retval 0 - on succes
  @retval 1 - error parsing expression in case pars function type
  @retval 2 - number of expression in case pars_set function type is <= 0
  @retval 3 - error parsing expression in case pars_set function type
  @retval 4 - unknown type of function
*/
long gfunct::read(XFILE *in)
{
  long i, j;
  Parser par;

  xfscanf (in,"%k%m","funct_type", &generalfunct_kwdset, (int*)&tfunc);

  switch (tfunc){
  case stat:{
    xfscanf (in,"%k%lf","const_val", &f);
    break;
  }
  case pars:{
    neqs = 1;
    eq   = new Equation* [neqs];
    var  = new variable**[neqs];
    func = new char*     [neqs];
    for (i = 0; i < neqs; i++)
      {
        func[i]=new char [256];
        xfscanf(in, "%k%255s", "func_formula", func[i]);
        //fgets(func[i], 255, in->file);
        eq[i] = par.TextToTree(func[i]);
        if (eq[i] == NULL)
	  {
	    print_err("Error parsing expression",__FILE__,__LINE__,__func__);
	    return (1);
	  }
        var[i] = new variable* [eq[i]->Variables.count()];
        for(j=0; j < eq[i]->Variables.count(); j++)
          var[i][j] = eq[i]->Variables.at(j);
      }
    break;
  }
  case tab:{
    tabf = new tablefunct;
    tabf->read(in);
    break;
  }
  case pars_set:{
    xfscanf (in, "%k%ld", "num_funct", &neqs);
    if (neqs <= 0)
      {
	print_err("Number of expressions in set <= 0",__FILE__,__LINE__,__func__);
        return (2);
      }
    eq     = new Equation* [neqs];
    var    = new variable**[neqs];
    func   = new char*     [neqs];
    limval = new double    [neqs];
    for (i = 0; i < neqs; i++)
      {
        xfscanf(in, "%k%le", "limval", limval+i);
        func[i]=new char [256];
        xfscanf(in, "%k%255s", "func_formula", func[i]);
        //fgets(func[i], 255, in->file);
        eq[i] = par.TextToTree(func[i]);
        if (eq[i] == NULL)
	  {
	    print_err("parsing expression", __FILE__, __LINE__, __func__);
	    return (3);
	}
        var[i] = new variable* [eq[i]->Variables.count()];
        for(j=0; j < eq[i]->Variables.count(); j++)
          var[i][j] = eq[i]->Variables.at(j);
      }
    break;
  }
  case itab:{
    itabf = new itablefunct;
    itabf->read(in);
    break;
  }
  default:{
    print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
    return (4);
  }
  }
  return(0);
}

/**
  This function computes value of given funtion expression for variable whose
  value is given by the t parameter.
  Parameters :
  @param t - given value of function variable
  Returns :
  Function value at parameter t.
*/
double gfunct::getval(double t)
{
  long i;
  double ret = 0.0;

  switch (tfunc)
  {
    case stat:
      ret=f;
      break;
    case pars:
      if (var[0][0])
        var[0][0]->Value = t;
      ret = eq[0]->Evaluate();
      break;
    case tab:
      ret=tabf[0].getval(t);
      break;
    case pars_set :
    {
      for (i = 0; i < neqs; i++)
      {
        if (limval[i] >= t)
        {
          if (var[i][0])
            var[i][0]->Value = t;
          ret = eq[i]->Evaluate();
          break;
        }
        if (i == neqs-1)
        {
          if (var[i][0])
            var[i][0]->Value = t;
          ret = eq[i]->Evaluate();
        }
      }
      break;
    }
    default:
    {
      print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
    }
  }
  
  return(ret);
}



/**
  This function computes value of given funtion expression for variable whose
  value is given by the t parameter.
  Parameters :
  @param t0 - given value of the first variable of function
  @param t  - given value of the second auxiliar variable of function
  Returns :
  Function value at parameter t.
*/
double gfunct::getval(double t, double t0)
{
  long i;
  double ret = 0.0;

 
  switch (tfunc)
  {
    case stat:
      ret=f;
      break;
    case pars:
      if (eq[0]->Variables.count() < 2)
      {
        getval(t);
        break;
      }
      if (var[0][0])
        var[0][0]->Value = t;
      if (var[0][1])
        var[0][1]->Value = t0;
      ret = eq[0]->Evaluate();
      break;
    case tab:
      ret=tabf[0].getval(t);
      break;
    case pars_set :
    {
      for (i = 0; i < neqs; i++)
      {
        if (limval[i] >= t)
        {
          if (eq[i]->Variables.count() < 2)
          {
            getval(t);
            break;
          }
          if (var[i][0])
            var[i][0]->Value = t;
          if (var[i][1])
            var[i][1]->Value = t0;
          ret = eq[i]->Evaluate();
          break;
        }
        if (i == neqs-1)
        {
          if (eq[i]->Variables.count() < 2)
          {
            getval(t);
            break;
          }
          if (var[i][0])
            var[i][0]->Value = t;
          if (var[i][1])
            var[i][1]->Value = t;
          ret = eq[i]->Evaluate();
        }
      }
      break;
    }
  default:
    {
      print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
    }
  }

  return(ret);
}



long gfunct::getval_long(double t)
{
  long ret;
  
  switch (tfunc){
  case itab:{
    ret=itabf[0].getval(t);
    break;
  }
  default:{
    print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
  }
  }
  
  return(ret);
}

/**
  This function reads data from the text file given by the parameter in. It is
  used in the mechprep preprocessor.
  Parameters :
  @param in - pointer to opened text file from which data will be read.
  Return :
  @retval 0 - on succes
  @retval 1 - error parsing expression in case pars function type
  @retval 2 - number of expression in case pars_set function type is <= 0
  @retval 3 - error parsing expression in case pars_set function type
  @retval 4 - unknown type of function
*/
long gfunct::read_prop(FILE *in)
{
  long i;
  Parser par;

  getint(in, (int &)tfunc);
  switch (tfunc)
  {
    case stat:
      getdouble(in, f);
      break;
    case pars:
      neqs = 1;
      eq   = new Equation* [neqs];
      func = new char*     [neqs];
      for (i = 0; i < neqs; i++)
      {
        func[i]=new char [256];
        getstring(in, func[i], 255);
        eq[i] = par.TextToTree(func[i]);
        if (eq[i] == NULL)
	  {
	    print_err("Error parsing expression", __FILE__, __LINE__, __func__);
	    return(1);
	  }
      }
      break;
  case tab:
      tabf = new tablefunct;
      tabf->read_prop(in);
      break;
    case pars_set :
    {
      getlong(in, neqs);
      if (neqs <= 0)
      {
	print_err("Number of expressions in set <= 0", __FILE__, __LINE__, __func__);
        return (2);
      }
      eq     = new Equation* [neqs];
      func   = new char*     [neqs];
      limval = new double    [neqs];
      for (i = 0; i < neqs; i++)
      {
        getdouble(in, limval[i]);
        func[i]=new char [256];
        getstring(in, func[i], 255);
        eq[i] = par.TextToTree(func[i]);
        if (eq[i] == NULL)
        {
	  print_err("Error parsing expression", __FILE__, __LINE__, __func__);
          return(3);
        }
      }
      break;
    }
    case itab:
      itabf = new itablefunct;
      itabf->read_prop(in);
    default:
      print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
      return(4);
  }
  return(0);
}

/**
  This function writes data to the text file given by the parameter in. It is
  used in the mechprep preprocessor.
  Parameters :
  @param out - pointer to opened text file from which data will be written.
  Return :
  @retval 0 - on succes
  @retval 1 - error parsing expression in case pars function type
  @retval 2 - number of expression in case pars_set function type is <= 0
  @retval 3 - error parsing expression in case pars_set function type
  @retval 4 - unknown type of function
*/
long gfunct::print(FILE *out)
{
  Parser par;
  long i;

  fprintf (out,"%d ",tfunc);

  switch (tfunc){
  case stat:{
    fprintf (out,"%f\n",f);
    break;
  }
  case pars:{
    fprintf(out, " %s\n", func[0]);
    break;
  }
  case tab:{
    tabf->print(out);
    break;
  }
  case pars_set:{
    fprintf (out, "%ld\n", neqs);
    for (i=0; i<neqs; i++)
      fprintf(out, " %s\n", func[i]);
    break;
  }
  case itab:{
    itabf->print(out);
    break;
  }
  default:{
    print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
    return (4);
  }
  }
  fprintf (out,"\n");

  return(0);
}

/**
   function initiates general function from gf
   
   JK
*/
void gfunct::initiate (gfunct &gf)
{
  long i;

  tfunc = gf.tfunc;
  
  switch (tfunc){
  case stat:{
    f=gf.f;
    break;
  }
  case tab:{
    tabf = new tablefunct;
    tabf->itype = gf.tabf->itype;
    tabf->asize = gf.tabf->asize;
    tabf->x=new double[tabf->asize];
    tabf->y=new double[tabf->asize];
    for (i=0;i<tabf->asize;i++){
      tabf->x[i]=gf.tabf->x[i];
      tabf->y[i]=gf.tabf->y[i];
    }
    break;
  }
  default:{
    print_err("unknown type of general function is required", __FILE__, __LINE__, __func__);
  }
  }
}

/**
   function initiates general function of table type
   
   @param nr - number of rows
   
   JK, 13.6.2005
*/
void gfunct::init_tab (long nr)
{
  tfunc = tab;
  tabf = new tablefunct;
  tabf->itype = piecewiselin;
  tabf->asize = nr;
  tabf->x = new double [nr];
  tabf->y = new double [nr];
}

/**
   function copies all data from subroutine argument to the actual object
   
   JK, 15.11.2008
*/
void gfunct::copy (gfunct &gf)
{
  //  type of general function
  tfunc = gf.tfunc;
  // number of expression sets
  neqs = gf.neqs;
  
  switch (tfunc){
  case stat:{
    //  constant value
    f = gf.f;
    break;
  }
  case tab:{
    tabf->copy (gf.tabf[0]);
    break;
  }
  case itab:{
    itabf->copy (gf.itabf[0]);
    break;
  }
  default:{
    print_err("unknown type of general function is required",__FILE__,__LINE__,__func__);
  }
  }
  
}
