#include "import_T3d.h"
#include "siftop_element_types.h"
#include <stdlib.h>
#include <stdio.h>

siftop *S=NULL ; ///< Pointer to siftop structure, which will contain topology from t3d file

/**
  Function sets topology pointer of global variable S to value oS.

   @param oS structure of siftop

   @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_set_topology_pointer ( siftop &oS )
{
    S=&oS ;
    return 0 ;
}

FILE *f ;

/**
  Function skips one line in file f

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void clln ( void )
{
/*    char dummy[1024] ;
    fgets( dummy,1024,f ) ;*/
  int c = 0;
  while ((! feof(f)) && ((c = fgetc(f)) != '\n'));
}

/**
  Function reads one integer vlaue from file f

  @param ovalue - result value

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void get_from_file ( long &ovalue )
{
    long j ;
    j=fscanf( f,"%ld",&ovalue ) ;
    if ( j==-1 )
    {
      fprintf(stderr, "\n\nError occured while reading integer number from file.\n" ) ;
      fprintf(stderr, " function get_from_file(long) (import_T3d.cpp)" ) ;
    }
}

/**
  Function reads one real value from file f

  @param ovalue - result value

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void get_from_file ( double &ovalue )
{
    long j ;
    j=fscanf( f,"%lf",&ovalue ) ;
    if ( j==-1 )
    {
      fprintf(stderr, "\n\nError occured while reading real number from file.\n" ) ;
      fprintf(stderr, " function get_from_file(double) (import_T3d.cpp)" ) ;
    }
}

long type,degree,renum,mode ;
long nodes,edges,trias,quads,tetras,pyrams,vedges,bricks,elements ;
long modeI[6] ;
long nodeI[20] ;

#define Vertex 1
#define Curve 2
#define Surface 3
#define Region 4
#define Patch 5
#define Shell 6

#define Parameters 0
#define Tangent 1
#define Normal 2
#define BoundaryEntities 3
#define AssociatedElements 4
#define Neighbourhood 5

/**
  This function decodes mode variable and stores results in modeI array

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void decode_mode ( void )
{
    long i ;
    for ( i=0 ; i<6 ; i++ )
    {
        modeI[i]=( mode>>i ) & 1 ;
    }
}

/**
  This function imports node coordinates and property number from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_nodes ( long paral )
{
    long i,id,property,entity,eid ;
    double x,y,z ;

//    for ( i=0 ; i<nodes-1 ; i++ )
    if (paral)
      S->aloc_pnn();
    for ( i=0 ; i<nodes ; i++ )
    {
        get_from_file( id ) ;
        if (paral)
          get_from_file(S->Pn[id-1]);
        get_from_file( x ) ;
        get_from_file( y ) ;
        get_from_file( z ) ;
        get_from_file( entity ) ;
        get_from_file( eid ) ;
        get_from_file( property ) ;

        S->set_node_coordinates( id-1,x,y,z ) ;
        S->set_node_properties( id-1,property ) ;

        switch ( entity )
        {
            case Vertex:
            case Region:
                break ;
            case Curve:
                if ( modeI[Parameters] || modeI[Tangent] )
                {
                    clln() ;
                }
                break ;
            case Surface:
            case Patch:
            case Shell:
                if ( modeI[Parameters] || modeI[Normal] )
                {
                    clln() ;
                }
                break ;
        }
    }
    return(0);
}

/**
  This function imports 1D elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_edges ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=ISOLinear1D ;
            esize=2 ;
            break ;
        case 2:
            etype=ISOQuadratic1D ;
            esize=3 ;
            break ;
    }

    for ( i=0 ; i<edges ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;

//        S->set_element_type( id,etype ) ;
//        S->set_element_nodes( id,esize,nodeI ) ;
//        S->set_element_properties( id,property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
    }
    return(0);
}

/**
  This function imports 2D triangle elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_trias ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=TriangleLinear ;
            esize=3 ;
            break ;
        case 2:
            etype=TriangleQuadratic ;
            esize=6 ;
            break ;
    }

    for ( i=0 ; i<trias ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;

//        S->set_element_type( id,etype ) ;
//        S->set_element_nodes( id,esize,nodeI ) ;
//        S->set_element_properties( id,property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
        S->set_element_nedges(id-1, 3);

        if ( modeI[Neighbourhood] )
        {
          clln() ;
        }
        if (modeI[BoundaryEntities])
        {
          for (j = 0; j < S->E[id-1].ned; j++)
          {
            get_from_file(dummy);
          }
          for (j = 0; j < S->E[id-1].ned; j++)
          {
            get_from_file(S->E[id-1].propedg[j]);
          }
        }
        else
          clln() ;
    }
    return(0);
}

/**
  This function imports 2D quad elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_quads ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=ISOLinear2D ;
            esize=4 ;
            break ;
        case 2:
            etype=ISOQuadratic2D ;
            esize=8 ;
            break ;
    }

    for ( i=0 ; i<quads ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;

//        S->set_element_type( id,etype ) ;
//        S->set_element_nodes( id,esize,nodeI ) ;
//        S->set_element_properties( id,property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
        S->set_element_nedges(id-1, 4);

/*        if ( modeI[BoundaryEntities] || modeI[Neighbourhood] )
        {
            clln() ;
        }*/
        if ( modeI[Neighbourhood] )
        {
          clln() ;
        }
        if (modeI[BoundaryEntities])
        {
          for (j = 0; j < S->E[id-1].ned; j++)
          {
            get_from_file(dummy);
          }
          for (j = 0; j < S->E[id-1].ned; j++)
          {
            get_from_file(S->E[id-1].propedg[j]);
          }
        }
        else
          clln() ;
    }
    return(0);
}

/**
  This function imports 3D tetrahedron elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
   modified 10.11.2005, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_tetras ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=TetrahedronLinear ;
            esize=4 ;
            break ;
        case 2:
            etype=TetrahedronQuadratic ;
            esize=10 ;
            break ;
    }

    for ( i=0 ; i<tetras ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
        S->set_element_nedges(id-1, 6);
        S->set_element_nsurf(id-1, 4);

        if ( modeI[Neighbourhood] || modeI[AssociatedElements] )
        {
            clln() ;
        }
        if (modeI[BoundaryEntities])
        {
          for (j = 0; j < S->E[id-1].nsurf; j++)
          {
            get_from_file(dummy);
            get_from_file(dummy);
          }
          for (j = 0; j < S->E[id-1].nsurf; j++)
          {
            get_from_file(S->E[id-1].propsurf[j]);
          }
          clln();
        }
        else
          clln();
       

    }
    return(0);
}

/**
  This function imports 3D pyramidal elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_pyrams ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=PyramideLinear ;
            esize=5 ;
            break ;
        case 2:
            etype=PyramideQuadratic ;
            esize=13 ;
            break ;
    }

    for ( i=0 ; i<pyrams ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;

//        S->set_element_type( id,etype ) ;
//        S->set_element_nodes( id,esize,nodeI ) ;
//        S->set_element_properties( id,property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
        S->set_element_nedges(id-1, 8);

        if ( modeI[BoundaryEntities] || modeI[Neighbourhood] || modeI[AssociatedElements] )
        {
            clln() ;
        }
/*        if ( modeI[Neighbourhood] || modeI[AssociatedElements] )
        {
            clln() ;
        }
        if (modeI[BoundaryEntities])
        {
          for (j = 0; j < S->E[id-1].ned; j++)
          {
            get_from_file(S->E[id-1].propedg[j]);
          }
        }*/
    }
    return(0);
}

/**
  This function imports 3D vedge elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d_vedges ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=VedgeLinear ;
            esize=6 ;
            break ;
        case 2:
            etype=VedgeQuadratic ;
            esize=15 ;
            break ;
    }

    for ( i=0 ; i<vedges ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;

//        S->set_element_type( id,etype ) ;
//        S->set_element_nodes( id,esize,nodeI ) ;
//        S->set_element_properties( id,property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
        S->set_element_nedges(id-1, 9);

        if ( modeI[BoundaryEntities] || modeI[Neighbourhood] || modeI[AssociatedElements] )
        {
            clln() ;
        }
/*        if ( modeI[Neighbourhood] || modeI[AssociatedElements] )
        {
            clln() ;
        }
        if (modeI[BoundaryEntities])
        {
          for (j = 0; j < S->E[id-1].ned; j++)
          {
            get_from_file(S->E[id-1].propedg[j]);
          }
        }*/
    }
    return(0);
}

/**
  This function imports 3D brick elements from t3d file f.
  Results are stored in S variable.

  @retval 0 always

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz

*/
long import_T3d_bricks ( void )
{
    long id,etype,esize,i,j,dummy,property ;
    switch ( degree )
    {
        case 1:
            etype=ISOLinear3D ;
            esize=8 ;
            break ;
        case 2:
            etype=ISOQuadratic3D ;
            esize=20 ;
            break ;
    }

    for ( i=0 ; i<bricks ; i++ )
    {
        get_from_file( id ) ;
        for ( j=0 ; j<esize ; j++ )
        {
            get_from_file( nodeI[j] ) ;
        }
        get_from_file( dummy ) ;
        get_from_file( dummy ) ;
        get_from_file( property ) ;

//        S->set_element_type( id,etype ) ;
//        S->set_element_nodes( id,esize,nodeI ) ;
//        S->set_element_properties( id,property ) ;
        S->set_element_type( id-1,etype ) ;
        S->set_element_nodes( id-1,esize,nodeI ) ;
        S->set_element_properties( id-1,property ) ;
        S->set_element_nedges(id-1, 12);
        S->set_element_nsurf(id-1, 6);

        if ( modeI[Neighbourhood] || modeI[AssociatedElements] )
        {
          clln() ;
        }

        if (modeI[BoundaryEntities])
        {
          for (j = 0; j < S->E[id-1].nsurf; j++)
          {
            get_from_file(dummy);
            get_from_file(dummy);
          }
          for (j = 0; j < S->E[id-1].nsurf; j++)
          {
            get_from_file(S->E[id-1].propsurf[j]);
          }
        }
        else
          clln() ;
    }
    return(0);
}

/**
  This function imports 3D pyramidal elements from t3d file f.
  Results are stored in S variable.

  @param ofname - string with valid name and path of T3d output file

  @retval -1 if topology pointer S is not set
  @retval -2 if file specified by ofname cannot be opened
  @retval  1

   created  29.5.2000, Ondrej Hrstka
   modified 23.1.2002, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long import_T3d ( char *ofname, long paral )
{
  long temp;
  
    if ( !S )
    {
        fprintf(stderr, "\n\nError - topology pointer not set.\n" ) ;
        fprintf(stderr, " function import_T3d() (import_T3d.cpp)\n" ) ;
        return -1 ;
    }
    f=fopen( ofname,"rt" ) ;
    if ( !f )
    {
        fprintf(stderr, "\n\nError - cannot open T3d output file %s for reading.\n",ofname ) ;
        fprintf(stderr, " function import_T3d() (import_T3d.cpp)\n" ) ;
        return -2 ;
    }

    get_from_file( type ) ;   // reads type of generator
    get_from_file( degree ) ; // reads degrees of element aproximation
    get_from_file( renum ) ;  // reads indicator of renumbering
    get_from_file( mode ) ;   // reads mode for elements
    
    if (paral==1){
      get_from_file( temp ) ;
    }
    
    decode_mode() ;

    get_from_file( nodes ) ;
//    nodes++ ;

    switch ( type )
    {
        case 3:
            get_from_file( edges ) ;
            get_from_file( trias ) ;
            get_from_file( tetras ) ;
            quads=0 ;
            pyrams=0 ;
            vedges=0 ;
            bricks=0 ;
            break ;
        case 4:
            get_from_file( edges ) ;
            get_from_file( quads ) ;
            get_from_file( bricks ) ;
            trias=0 ;
            tetras=0 ;
            pyrams=0 ;
            vedges=0 ;
            break ;
        case 7:
            get_from_file( edges ) ;
            get_from_file( trias ) ;
            get_from_file( quads ) ;
            get_from_file( tetras ) ;
            get_from_file( pyrams ) ;
            get_from_file( vedges ) ;
            get_from_file( bricks ) ;
            break ;
    }

    S->set_number_of_nodes( nodes ) ;
    import_T3d_nodes(paral) ;

//    elements=edges+trias+quads+tetras+pyrams+vedges+bricks+1 ;
    elements=edges+trias+quads+tetras+pyrams+vedges+bricks ;
    S->set_number_of_elements( elements ) ;

    import_T3d_edges() ;
    import_T3d_trias() ;
    import_T3d_quads() ;
    import_T3d_tetras() ;
    import_T3d_pyrams() ;
    import_T3d_vedges() ;
    import_T3d_bricks() ;

    return 1 ;
}










