#include "seqfeti.h"
#include "gmatrix.h"

seqfeti::seqfeti ()
{
  zero=1.0e-100;
  
  //  number of subdomains
  ns=0;
  
  //  numbers of nodes on subdomains
  //nnsd=NULL;
  
  //  dimensions of the kernels - numbers of rigid body modes
  nrbmdom=NULL;
  //  list of linearly dependent equations
  se=NULL;
  //  array for rigid body modes on subdomains (twodimensional array)
  rbmdom=NULL;
  
  
  


  //  node-subdomain correspondence
  nsid = NULL;
  //  numbers of DOFs on subdomains
  ndofdom = NULL;
  //  list of DOFs on subdomains
  cndom = NULL;

  //  first node numbers on subdomains
  //fnnsd=NULL;
  
  wscalmat = NULL;
  lwscalmat = NULL;
    
    
  smsky = NULL;
  smscr = NULL;

}

seqfeti::~seqfeti ()
{
  long i;

  for (i=0;i<ns;i++){
    delete [] se[i];
    delete [] lwscalmat[i];
  }
  delete [] se;
  delete [] wscalmat;
  delete [] lwscalmat;
  
  for (i=0;i<ns;i++){
    delete [] rbmdom[i];
  }
  delete [] rbmdom;

  
  





  
  delete [] nsid;
  delete [] ndofdom;
  

  //delete [] fnnsd;

  for (i=0;i<ns;i++){
    delete [] cndom[i];
  }
  delete [] cndom;
}

/**
   function reads basic data
   
   @param in - input file
   
   JK, 20.11.2007
*/
void seqfeti::read (XFILE *in)
{
  //  number of subdomains
  //  number of estimated rigid body modes on one subdomain
  //  threshold for zero detection on the diagonal (threshold for detection of dependent row)
  //  maximum number of iterations in the conjugate gradient method
  //  required norm of residual
  xfscanf (in,"%ld %ld %le %ld %le",&ns,&ense,&thresh,&nicg,&errcg);
  //  type of preconditioner
  xfscanf (in,"%d",(int*)&prec);
}

/**
   function prints basic data
   
   @param out - output file
   
   JK, 24.11.2008
*/
void seqfeti::print (FILE *out)
{
  //  number of subdomains
  //  number of estimated rigid body modes on one subdomain
  //  threshold for zero detection on the diagonal (threshold for detection of dependent row)
  //  maximum number of iterations in the conjugate gradient method
  //  required norm of residual
  fprintf (out,"%ld %ld %le %ld %le\n",ns,ense,thresh,nicg,errcg);
  //  type of preconditioner
  fprintf (out,"%d\n",prec);
}



/**
   JK, 17.9.2007
*/
void seqfeti::initiate (seqselnodes *selnodfeti,gtopology *top,FILE *out)
{
  long i,j,k,ndofn,dof,nid;
  long *red;
  red = new long [ns+1];
  
  //  number of DOFs (unknowns) in coarse problem = total number of boundary DOFs
  ndofcp =selnodfeti->tndof;
  
  ncdofd = new long [ns];
  for (i=0;i<ns;i++){
    ncdofd[i] = selnodfeti->ndofdom[i];
  }
  
  // scaling matrix
  wscalmat = new double [ndofcp]; 
  for (i = 0; i < ndofcp; i++){
    wscalmat[i] = double(selnodfeti->dofmultip[i]);
  }
 
  //  array containing code numbers contributing to the coarse problem
  //edofs = selnodfeti->ldof;
  edofs = new long* [ns];
  for (i=0;i<ns;i++){
    edofs[i] = new long [ncdofd[i]];
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<ncdofd[i];j++){
      edofs[i][j] = selnodfeti->ldof[i][j];
    }
  }

  red[0]=0;
  nid=0;
  for (i=0;i<ns;i++){
    red[i+1]=0;
    for (j=0;j<top->stop->nnsd[i];j++){
      ndofn = top->give_ndofn (nid);
      for (k=0;k<ndofn;k++){
	dof = top->give_dof (nid,k);
	if (dof>red[i+1]){
	  red[i+1]=dof;
	}
      }
      nid++;
    }
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<ncdofd[i];j++){
      if (edofs[i][j]>0)
	edofs[i][j] -= red[i];
      else
	edofs[i][j] += red[i];
    }
  }

  fprintf (out,"\n\n\n osel \n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n");
    for (j=0;j<ncdofd[i];j++){
      fprintf (out,"\n edofs %ld",edofs[i][j]);
    }
  }
  fprintf (out,"\n\n\n osel \n");








  ccn = new long* [ns];
  for (i=0;i<ns;i++){
    ccn[i] = new long [ncdofd[i]];
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<ncdofd[i];j++){
      ccn[i][j] = selnodfeti->cndom[i][j];
    }
  }
  
  delete [] red;
  /*
  // lokalni wscalmat
  lwscalmat = new double* [ns]; 
  nlwscalmat = new long [ns];
  for (i = 0; i < ns; i++){
    nlwscalmat[i] = selnodfeti->ndofdom[i];
    lwscalmat[i] = new double [ nlwscalmat[i]]; 
    for (j = 0; j <  nlwscalmat[i]; j++){
      lwscalmat[i][j] = double(selnodfeti->ldofmultip[i][j]);
    }
  }
  */
}



/**
   function orders unknowns on subdomains with respect to the FETI method
   
   @param top - pointer to domain topology
   
   JK, 31.7.2005
*/
void seqfeti::subdomain_ordering (gtopology *top,FILE *out)
{
  ndof = top->codenum_generation (out);
}


/**
   function determines ndofmax
   the variable is used for simple allocation in other functions
   
   JK, 16.9.2007
*/
void seqfeti::det_ndofmax ()
{
  long i;
  
  ndofmax=0;
  for (i=0;i<ns;i++){
    if (ndofmax<ndofdom[i])
      ndofmax=ndofdom[i];
  }
  
  //for (i=0;i<ns;i++){
  //if (ndofmax<ncdofd[i])
  //ndofmax=ncdofd[i];
  //}
}

/**
   function assembles list of unknowns belonging to the subdomains
   
   @param top - pointer to topology
   
   JK, 21.8.2007
*/
void seqfeti::assemble_subdom_unknowns (gtopology *top,FILE *out)
{
  long i,j,k,l,ndofn;
  
  nsid = new long [top->nn];
  l=0;
  for (i=0;i<ns;i++){
    for (j=0;j<top->stop->nnsd[i];j++){
      nsid[l]=i;
      l++;
    }
  }
  
  
  //  number of DOFs on subdomains
  ndofdom = new long [ns];
  for (i=0;i<ns;i++){
    ndofdom[i]=0;
  }
  
  for (i=0;i<top->nn;i++){
    ndofn=top->give_ndofn (i);
    l=nsid[i];
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if (k>0)
	ndofdom[l]++;
    }
  }
  
  cndom = new long* [ns];
  for (i=0;i<ns;i++){
    cndom[i] = new long [ndofdom[i]];
    ndofdom[i]=0;
  }
  
  for (i=0;i<top->nn;i++){
    ndofn=top->give_ndofn (i);
    l=nsid[i];
    for (j=0;j<ndofn;j++){
      k=top->give_dof (i,j);
      if (k>0){
	cndom[l][ndofdom[l]]=k;
	ndofdom[l]++;
      }
    }
  }
  
  fprintf (out,"\n\n kontrola pole ndofdom");
  for (i=0;i<ns;i++){
    fprintf (out,"\n ndofdom %6ld   %6ld",i,ndofdom[i]);
  }
  fprintf (out,"\n\n kontrola pole cndom");
  for (i=0;i<ns;i++){
    fprintf (out,"\n domain %ld",i);
    for (j=0;j<ndofdom[i];j++){
      fprintf (out,"\n dom %6ld  dof %6ld   %6ld",i,j,cndom[i][j]);
    }
  }  
  


  jum = new double* [ns];
  for (i=0;i<ns;i++){
    jum[i] = new double [ndofdom[i]];
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<ndofdom[i];j++){
      jum[i][j]=0.0;
    }
  }

}


/**
   function assembles subdomain matrices from the %matrix of the whole system
   
   @param gm - pointer to the %matrix of the system
   
   JK, 21.8.2007
*/
void seqfeti::subdomain_matrices (gmatrix *gm,FILE *out)
{
  long i,j,k;
  
  smsky = new skyline [ns];
  for (i=0;i<ns;i++){
    smsky[i].assemble_from_scr (gm->scr->adr,gm->scr->ci,gm->scr->a,ndofdom[i],cndom[i]);
  }
  
  long *aux;
  ndofprec = new long [ns];
  cnprec = new long* [ns];
  cpreccn = new long* [ns];

  for (i=0;i<ns;i++){
    aux = new long [ndofdom[i]];
    for (j=0;j<ndofdom[i];j++){
      aux[j]=0;
    }
    
    for (j=0;j<ncdofd[i];j++){
      k=edofs[i][j];
      if (k<0)
	k=0-k-1;
      else
	k--;
      if (k>-1)
	aux[k]++;
    }
    
    ndofprec[i]=0;
    for (j=0;j<ndofdom[i];j++){
      if (aux[j]>0)
	ndofprec[i]++;
    }
    
    fprintf (out,"\n\n aux  domena %ld\n",i);
    for (j=0;j<ndofdom[i];j++){
      fprintf (out," %ld",aux[j]);
    }

    
    cnprec[i] = new long [ndofprec[i]];
    cpreccn[i] = new long [ndofprec[i]];
    k=0;
    for (j=0;j<ndofdom[i];j++){
      if (aux[j]>0){
	cnprec[i][k]=cndom[i][j];
	cpreccn[i][k]=j;
	k++;
      }
    }
    
    delete [] aux;
  }
  
  fprintf (out,"\n\n\n\n kontrola ve funkci subdomain_matrices (file %s, line %d)\n\n",__FILE__,__LINE__);
  for (i=0;i<ns;i++){
    fprintf (out,"\n domena %5ld     ndofprec %ld\n",i,ndofprec[i]);
    for (j=0;j<ndofprec[i];j++){
      fprintf (out,"  %ld",cnprec[i][j]);
    }
    fprintf (out,"\n");
    for (j=0;j<ndofprec[i];j++){
      fprintf (out,"  %ld",cpreccn[i][j]);
    }
  }
  fprintf (out,"\n\n\n\n\n");


  if (prec==feti_lumped){
    smscr = new symcomprow [ns];
    for (i=0;i<ns;i++){
      //gm->scr->select_submatrix (&smscr[i],ndofdom[i],cndom[i]);
      gm->scr->select_submatrix (&smscr[i],ndofprec[i],cnprec[i]);
    }
  }
  
  if (prec==feti_dirichlet){
    smdm = new densemat [ns];

    double *x,*y,*condvect;
    for (i=0;i<ns;i++){
      x = new double [ndofprec[i]];
      y = new double [ndofprec[i]];
      smdm[i].a = new double [ndofprec[i]*ndofprec[i]];
      smdm[i].negm=ndofprec[i]*ndofprec[i];
      smdm[i].n=ndofprec[i];
      condvect=new double [ndofprec[i]];

      smsky[i].ldlkon_sky (smdm[i].a,condvect,x,y,ndofprec[i],1);
      smsky[i].~skyline ();
      
      delete [] x;
      delete [] y;
      delete [] condvect;
    }
    
    for (i=0;i<ns;i++){
      smsky[i].assemble_from_scr (gm->scr->adr,gm->scr->ci,gm->scr->a,ndofdom[i],cndom[i]);
    }
    
  }
  
  //fclose (out);
  //abort ();

  /*
  delete [] ndofprec;
  for (i=0;i<ns;i++){
    delete [] cnprec[i];
  }
  delete [] cnprec;
  */
}

/**
   function computes kernel (rigid body modes) of the %matrix
   
   JK, 28.8.2007
*/
void seqfeti::kernel (FILE *out)
{
  long i,j;
  
  //  array for numbers of rigid body modes
  if (nrbmdom!=NULL){
    delete [] nrbmdom;
  }
  nrbmdom = new long [ns];
  
  //  array for linearly dependent equation numbers
  if (se!=NULL){
    for (i=0;i<ns;i++){
      delete [] se[i];
    }
    delete [] se;
  }
  se = new long* [ns];
  for (i=0;i<ns;i++){
    se[i] = new long [ense];
    for (j=0;j<ense;j++){
      se[i][j]=-1;
    }
  }
  
  //  array for rigid body modes
  if (rbmdom!=NULL){
    for (i=0;i<ns;i++){
      delete [] rbmdom[i];
    }
    delete [] rbmdom;
  }
  rbmdom = new double* [ns];
  for (i=0;i<ns;i++){
    rbmdom[i] = new double [ense*ndofdom[i]];
    for (j=0;j<ense*ndofdom[i];j++){
      rbmdom[i][j]=0.0;
    }
  }
  
  
  //  loop over subdomains
  for (i=0;i<ns;i++){
    smsky[i].ker (rbmdom[i],nrbmdom[i],se[i],ense,thresh,3);

    //smsky[i].printmat (out);
  }
  
}
 




















/**
   function extracts contributions from local %vector to coarse %vector
   
   @param nd - domain id (number of required subdomain)
   @param lv - local %vector
   @param cv - coarse %vector
   
   JK, 14.9.2007
*/
void seqfeti::local_coarse (long nd,double *lv,double *cv)
{
  long i,j,k;
  
  for (i=0;i<ncdofd[nd];i++){
    j=edofs[nd][i];
    k=ccn[nd][i]-1;
    if (j<0){
      j=0-j-1;
      cv[k]-=lv[j];
    }
    else{
      j=j-1;
      cv[k]+=lv[j];
    }
  }
}


/**
   function extracts contributions from coarse %vector to local %vector
   
   @param nd - domain id (number of required subdomain)
   @param lv - local %vector
   @param cv - coarse %vector

   JK, 14.9.2007
*/
void seqfeti::coarse_local (long nd,double *lv,double *cv)
{
  long i,j,k;
  
  for (i=0;i<ncdofd[nd];i++){
    k=ccn[nd][i]-1;
    j=edofs[nd][i];
    if (j<0){
      j=0-j-1;
      lv[j]-=cv[k];
    }
    else{
      j=j-1;
      lv[j]+=cv[k];
    }
  }
}







/**
   function computes size of matrix H
   
   @param out - output stream
   
   JK, 16.9.2007
*/
void seqfeti::hmatrixsize (FILE *out)
{
  long i;
  
  //  addresses computed from nrbmdom array
  rbmadr = new long [ns+1];
  for (i=0;i<ns+1;i++){
    rbmadr[i]=0;
  }
  
  //  computation of addresses
  rbmadr[0]=0;
  for (i=0;i<ns;i++){
    rbmadr[i+1]=rbmadr[i]+nrbmdom[i];
  }
  
  //  size of the matrix H
  hsize=rbmadr[ns];
  
}



/**
   function assembles %matrix H
   columns are RBM after localization
   
   @param h - %matrix H
   @param rbm - array containing rigid body motions
   
   JK, 16.9.2007
*/
void seqfeti::hmatrix (double *h)
{
  long i,k,l;
  double *av;
  
  av = new double [ndofcp];
  
  for (i=0;i<ns;i++){
    for (k=0;k<nrbmdom[i];k++){
      nullv (av,ndofcp);
      local_coarse (i,rbmdom[i]+k*ndofdom[i],av);
      for (l=0;l<ndofcp;l++){
	h[l*hsize+rbmadr[i]+k]=0.0-av[l];
      }
    }
  }
  
  delete [] av;
}


/**
   function assembles vector q
   
   @param q - vector q
   @param rbm - array containing rigid body motions
   @param f - right hand side
   
   JK, 16.9.2007
*/
void seqfeti::qvector (double *q,double **ff)
{
  long i,j,k;
  
  k=0;
  for (i=0;i<ns;i++){
    for (j=0;j<nrbmdom[i];j++){
      q[k]=0.0-ss(&rbmdom[i][j*ndofdom[i]],ff[i],ndofdom[i]);
      k++;
    }
  }
  
}



/**
   function computes projection in FETI method
   v_new = v_old - H . (H^T . H)^{-1} . H^T . v_old
   function overwrites vector v by projected vector
   
   @param v - %vector
   @param h - matrix H
   @param h1 - matrix (H^T.H)^{-1}
   @param ndofcp,hsize - number of rows and columns of matrix H
   
   ndofcp - number of Lagrange multipliers
   hsize - total number of all rigid body motions
   
   JK, 16.9.2007
*/
void seqfeti::feti_projection (double *v,double *h,double *h1)
{
  long i;
  double *p,*q,*w;
  
  p = new double [hsize];
  q = new double [hsize];
  w = new double [ndofcp];

  //  p = H^T.v
  //mtvc (h,v,p,ndofcp,hsize);
  mtxv (h,v,p,ndofcp,hsize);
  
  
  /*
  fprintf (out,"\n\n\n kontrola H^T.v\n");
  for (i=0;i<hsize;i++){
    fprintf (out," %lf",p[i]);
  }
  */

  
  //  q = (H^T.H)^{-1}.p
  mxv (h1,p,q,hsize,hsize);
  

  /*
  fprintf (out,"\n\n\n kontrola (H^T.H)^{-1}.H^T.v\n");
  for (i=0;i<hsize;i++){
    fprintf (out," %lf",q[i]);
  }
  */


  //  w = H.q
  mxv (h,q,w,ndofcp,hsize);
  
  
  /*
  fprintf (out,"\n\n\n kontrola H.(H^T.H)^{-1}.H^T.v\n");
  for (i=0;i<ndofcp;i++){
    fprintf (out," %lf",w[i]);
  }
  */
  
  //  v_new = v_old - w
  for (i=0;i<ndofcp;i++){
    v[i]-=w[i];
  }
  
  delete [] w;  delete [] q;  delete [] p;
}


/**
   function obtaines jumps from problem solved
   
   @param jump - array of jumps
   
   JK, 28.11.2007
*/
void seqfeti::get_jumps (double *jump)
{
  long i,j;
  
  /*
  if (jum==NULL){
    jum = new double* [ns];
    for (i=0;i<ns;i++){
      jum[i] = new double [ndofdom[i]];
    }
  }
  */
  for (i=0;i<ns;i++){
    for (j=0;j<ndofdom[i];j++){
      jum[i][j]=0.0;
    }
  }
  
  
  //  allocation of vectors of right hand sides on subdomains
  for (i=0;i<ns;i++){
    globloc (jump,jum[i],cndom[i],ndofdom[i]);
  }
  
}



void seqfeti::scaling(double *invect,double *outvect,long n,FILE *out)
{
  double *aux;
  long i;
  aux = new double[n];
  
  for(i = 0; i < n; i++){
    aux[i] = invect[i]/wscalmat[i];
  }
  
  /*
  fprintf (out,"\n\n\n kontrola skalovani \n");
  for(i = 0; i < n; i++){
    fprintf (out,"%ld  pred %le po %le wscalmat %lf\n",i+1,invect[i],aux[i],wscalmat[i]); 
  }
  */
  
  for(i = 0; i < n; i++){
    outvect[i] = aux[i];
  }
  
  delete []aux;
}

void seqfeti::lscaling(double *invect,double *outvect,long ndom,FILE *out)
{
  double *aux;
  long i;
  
  aux = new double[nlwscalmat[ndom]];
  
  for(i = 0; i < nlwscalmat[ndom]; i++){
    aux[i] = invect[i]/lwscalmat[ndom][i];
  }
  
  
  fprintf (out,"\n\n\n kontrola skalovani \n");
  for(i = 0; i <  nlwscalmat[ndom]; i++){
    fprintf (out,"%ld  pred %le po %le wscalmat %lf ccn %ld\n",i+1,invect[i],aux[i],lwscalmat[ndom][i],ccn[ndom][i]); 
  }
  
  
  for(i = 0; i <  nlwscalmat[ndom]; i++){
    outvect[i] = aux[i];
  }
  
  delete []aux;
}



/**

*/
void seqfeti::lumpedprec (long nd,double *dd,double *pp)
{
  long i;
  double *d,*p;
  
  d = new double [ndofprec[nd]];
  p = new double [ndofprec[nd]];
  
  for (i=0;i<ndofprec[nd];i++){
    d[i]=0.0;
    p[i]=0.0;
  }
  
  for (i=0;i<ndofprec[nd];i++){
    d[i]=dd[cpreccn[nd][i]];
  }

  smscr[nd].mxv_scr (d,p);
  
  for (i=0;i<ndofprec[nd];i++){
    pp[cpreccn[nd][i]]=p[i];
  }
  
  delete [] p;
  delete [] d;
}

/**
   JK, 7.5.2008
*/
void seqfeti::dirichletprec (long nd,double *dd,double *pp)
{
  long i;
  double *d,*p;
  
  d = new double [ndofprec[nd]];
  p = new double [ndofprec[nd]];
  
  for (i=0;i<ndofprec[nd];i++){
    d[i]=0.0;
    p[i]=0.0;
  }
  
  for (i=0;i<ndofprec[nd];i++){
    d[i]=dd[cpreccn[nd][i]];
  }

  smdm[nd].mxv_dm (d,p);
  
  for (i=0;i<ndofprec[nd];i++){
    pp[cpreccn[nd][i]]=p[i];
  }
  
  delete [] p;
  delete [] d;
}



/**
   function performs modified conjugate gradient method
   
   @param w - vector of Lagrange multipliers
   @param ff - %vectors of right hand side
   @param h - matrix H
   @param h1 - matrix (H^T.H)^{-1}
   @param ndofcp,hsize - number of rows and columns of matrix H
   
   ndofcp - number of Lagrange multipliers
   hsize - total number of all rigid body motions
   
   JK, 16.9.2007
*/
void seqfeti::mpcg (gtopology *top,gmatrix *gm,
		    double *w,double **ff,double *q,double *h,double *h1,FILE *out)
{
  long i,j,k;
  double nom,denom,alpha,beta;
  double *d,*dd,*g,*p,*pp,*hq;

  // JB
  //prepare_scaling(top,out);
  
  //  vectors defined on each subdomain
  dd = new double [ndofmax];
  pp = new double [ndofmax];

  
  //  direction vector allocation
  d = new double [ndofcp];
  //  residuum vector allocation
  g = new double [ndofcp];
  //  auxiliary vector allocation
  p = new double [ndofcp];
  
  //fprintf (stdout,"\n\n\n ndofcp %ld       hsize  %ld",ndofcp,hsize);
  
  //  initiation of conjugate gradients
  
  hq = new double [hsize];
  //  (H^T.H)^{-1} q
  mxv (h1,q,hq,hsize,hsize);
  //  H [(H^T.H)^{-1} q]
  mxv (h,hq,w,ndofcp,hsize);
  delete [] hq;
  
  nullv (g,ndofcp);
  for (j=0;j<ns;j++){
    nullv (dd,ndofdom[j]);
    coarse_local (j,dd,w);
    // f - \lambda_0
    subv (dd,ff[j],ndofdom[j]);
    
    fprintf (out,"\n");
    for (i=0;i<ndofdom[j];i++){
      fprintf (out,"\n zoufalec %4ld    %20.15e",i,dd[i]);
    }

    //  K^+ (f - \lambda_0) = r
    smsky[j].ldl_feti_sky (pp,dd,nrbmdom[j],se[j],zero);
    
    fprintf (out,"\n");
    for (i=0;i<nrbmdom[j];i++){
      fprintf (out,"\n zoufalec    se %4ld",se[j][i]);
    }
    fprintf (out,"\n");
    for (i=0;i<ndofdom[j];i++){
      fprintf (out,"\n zoufalec %4ld    %20.15e",i,pp[i]);
    }
    
    //  zacatek - streda 28.11.2007
    //  odecet skoku
    subv (pp,jum[j],ndofdom[j]);
    //  konec - streda 28.11.2007
    
    
    //nullvr (buff,maxlggl+1);
    local_coarse (j,pp,g);
  }
  
  fprintf (out,"\n");
  for (i=0;i<ndofcp;i++){
    fprintf (out,"\n zoufalec %4ld    %20.15e",i,g[i]);
  }
  
  //  P r
  feti_projection (g,h,h1);
  
  //  initialization of direction vector
  for (i=0;i<ndofcp;i++){
    d[i]=0.0-g[i];
  }
  
  //  nominator evaluation
  nom = ss (g,g,ndofcp);
  
  
  
  // *************************
  //  main iteration loop   **
  // *************************
  for (i=0;i<nicg;i++){
    
    // ******************************************
    //   auxiliary vector computation K.d = p  **
    // ******************************************
    nullv (p,ndofcp);
    for (j=0;j<ns;j++){
      nullv (dd,ndofmax);
      coarse_local (j,dd,d);
      
      //  computation of K^+ d on master
      smsky[j].ldl_feti_sky (pp,dd,nrbmdom[j],se[j],zero);
      
      //nullvr (buff,maxlggl+1);
      local_coarse (j,pp,p);
    }
    
    //  denominator of alpha
    denom = ss (d,p,ndofcp);
    
    //fprintf (stdout,"\n denominator u alpha   %le",denom);
    
    fprintf (out,"\n\n kontrola citatele a jmenovatele pred alpha %e / %e",nom,denom);
    //fprintf (stderr,"\n\n kontrola citatele a jmenovatele pred alpha %lf / %lf",nom,denom);
    
    if (fabs(denom)<zero){
      fprintf (stderr,"\n\n zero denominator in modified conjugate gradient method (file %s, line %d).\n",__FILE__,__LINE__);
      break;
    }
    
    // *******************
    //   vypocet alpha  **
    // *******************
    alpha = nom/denom;
    
    // **************************************************************
    //  vypocet noveho gradientu g a nove aproximace neznamych x   **
    // **************************************************************
    for (j=0;j<ndofcp;j++){
      w[j]+=alpha*d[j];
      g[j]+=alpha*p[j];
    }
    
    feti_projection (g,h,h1);
    
    
    // *******************************************************
    // *******************************************************
    //  vlozka s predpodminenim
    // *******************************************************
    // *******************************************************
    
    nullv (p,ndofcp);
    
    switch (prec){
    case noprecond:{
      for (j=0;j<ns;j++){
	nullv (dd,ndofmax);
	coarse_local (j,dd,g);
	nullv (pp,ndofmax);
	for (k=0;k<ndofdom[j];k++){
	  pp[k]=dd[k];
	}
	local_coarse (j,pp,p);
      }
      break;
    }
    case feti_lumped:{
      scaling(g,g,ndofcp,out);
      for (j=0;j<ns;j++){
	nullv (dd,ndofmax);
	coarse_local (j,dd,g);
	//lscaling(dd,dd,j,out);
	nullv (pp,ndofmax);
	
	lumpedprec (j,dd,pp);
	
	//lscaling(pp,pp,j,out);
	local_coarse (j,pp,p);
      }
      scaling(p,p,ndofcp,out);
      break;
    }
    case feti_dirichlet:{
      scaling(g,g,ndofcp,out);
      for (j=0;j<ns;j++){
	nullv (dd,ndofmax);
	coarse_local (j,dd,g);
	//lscaling(dd,dd,j,out);
	nullv (pp,ndofmax);
	
	dirichletprec (j,dd,pp);
	
	//lscaling(pp,pp,j,out);
	local_coarse (j,pp,p);
      }
      scaling(p,p,ndofcp,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of preconditioner is required in function mpcg (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }
    
    feti_projection (p,h,h1);
    
    // *******************************************************
    // *******************************************************
    //   konec vlozky s predpodminenim
    // *******************************************************
    // *******************************************************
    
    
    denom = nom;
    if (fabs(denom)<zero){
      fprintf (stderr,"\n\n zero denominator of beta in function (file %s, line %d).\n",__FILE__,__LINE__);
      break;
    }
    
    //nom = ss (g,g,ndofcp);
    nom=ss(p,g,ndofcp);
    
    fprintf (stdout,"\n iteration  %4ld        norm g   %e",i,sqrt(nom));
    fprintf (out,"\n iteration  %4ld        norm g   %e",i,sqrt(nom));
    
    fprintf (out,"\n kontrola citatele a jmenovatele pred betou  %e / %e",nom,denom);
    //fprintf (stderr,"\n kontrola citatele a jmenovatele pred betou  %lf / %lf",nom,denom);
    
    if (sqrt(nom)<errcg){
      break;
    }
    
    
    //  computation of beta coefficient
    beta = nom/denom;
    
    //  new direction vector
    for (j=0;j<ndofcp;j++){
      d[j]=beta*d[j]-p[j];
    }
    
  }
  
  anicg=i;  aerrcg=nom;
  
  fprintf (out,"\n\n\n\n kontrola Lagrangeovych multiplikatoru \n");
  for (i=0;i<ndofcp;i++){
    fprintf (out,"\n lagr. mult %4ld    %e",i,w[i]);
  }
  
  delete [] d;
  delete [] p;
  delete [] g;
  
  delete [] dd;
  delete [] pp;
}


/**
   function performs modified preconditioned reorthonormalized conjugate gradient method
   
   @param w - vector of Lagrange multipliers
   @param ff - %vectors of right hand side
   @param h - matrix H
   @param h1 - matrix (H^T.H)^{-1}
   @param ndofcp,hsize - number of rows and columns of matrix H
   
   ndofcp - number of Lagrange multipliers
   hsize - total number of all rigid body motions
   
   JK, 5.5.2008
*/
void seqfeti::mprcg (gtopology *top,gmatrix *gm,
		     double *lambda,double **ff,double *e,double *g,double *g1,FILE *out)
{
  long i,j,k;
  double nom,denom,alpha,s;
  double *ge,*r,*lambdad,*rr,*w,*p,*y,*op,*fp,*pp,*fpd,*ofp,*coef,*wd,*yd,*beta;

    r = new double [ndofcp];
  lambdad = new double [ndofmax];
  rr = new double [ndofmax];
  w = new double [ndofcp];
  p = new double [ndofcp];
  y = new double [ndofcp];
  op = new double [ndofcp*101];
  fp = new double [ndofcp];
  pp = new double [ndofmax];
  fpd = new double [ndofmax];
  ofp = new double [ndofcp*101];
  coef = new double [101];
  wd = new double [ndofmax];
  yd = new double [ndofmax];
  beta = new double [101];


  //  vectors defined on each subdomain

  


  //  initiation of conjugate gradients
  
  ge = new double [hsize];
  //  (G^T.G)^{-1} e
  mxv (g1,e,ge,hsize,hsize);
  //  G [(G^T.G)^{-1} e]
  mxv (g,ge,lambda,ndofcp,hsize);
  delete [] ge;
  
  nullv (w,ndofcp);
  for (j=0;j<ns;j++){
    nullv (lambdad,ndofdom[j]);
    coarse_local (j,lambdad,lambda);
    // f - \lambda_0
    subv (lambdad,ff[j],ndofdom[j]);
    
    //  K^+ (f - \lambda_0) = r
    smsky[j].ldl_feti_sky (wd,lambdad,nrbmdom[j],se[j],zero);
    
    
    //  zacatek - streda 28.11.2007
    //  odecet skoku
    //subv (pp,jum[j],ndofdom[j]);
    //  konec - streda 28.11.2007
    
    
    //nullvr (buff,maxlggl+1);
    local_coarse (j,wd,w);
  }
  
  for (j=0;j<ndofcp;j++){
    w[j]*=-1.0;
  }

  //  P r
  feti_projection (w,g,g1);
  
  //  initialization of direction vector
  for (i=0;i<ndofcp;i++){
    p[i]=w[i];
    y[i]=w[i];
  }

  for (j=0;j<ndofcp;j++){
    op[0*ndofcp+j]=p[j];
  }
  
  
  // *************************
  //  main iteration loop   **
  // *************************
  for (i=0;i<nicg;i++){
    
    // ******************************************
    //   auxiliary vector computation K.d = p  **
    // ******************************************
    nullv (fp,ndofcp);
    for (j=0;j<ns;j++){
      nullv (pp,ndofmax);
      coarse_local (j,pp,p);
      
      //  computation of K^+ d on master
      smsky[j].ldl_feti_sky (fpd,pp,nrbmdom[j],se[j],zero);
      
      //nullvr (buff,maxlggl+1);
      local_coarse (j,fpd,fp);
    }
    
    for (j=0;j<ndofcp;j++){
      ofp[i*ndofcp+j]=fp[j];
    }
    
    nom = ss (p,w,ndofcp);

    //  denominator of alpha
    denom = ss (p,fp,ndofcp);
    
    coef[i]=denom;

    //fprintf (stdout,"\n denominator u alpha   %le",denom);
    
    fprintf (out,"\n\n kontrola citatele a jmenovatele pred alpha %e / %e",nom,denom);
    //fprintf (stderr,"\n\n kontrola citatele a jmenovatele pred alpha %lf / %lf",nom,denom);
    
    if (fabs(denom)<zero){
      fprintf (stderr,"\n\n zero denominator in modified conjugate gradient method (file %s, line %d).\n",__FILE__,__LINE__);
      break;
    }
    
    // *******************
    //   vypocet alpha  **
    // *******************
    alpha = nom/denom;
    
    // **************************************************************
    //  vypocet noveho gradientu g a nove aproximace neznamych x   **
    // **************************************************************
    nom=0.0;
    for (j=0;j<ndofcp;j++){
      lambda[j]+=alpha*p[j];
      w[j]-=alpha*fp[j];
      nom+=w[j]*w[j];
    }
    
    fprintf (stdout,"\n iteration  %4ld        norm g   %e",i,sqrt(nom));
    fprintf (out,"\n iteration  %4ld        norm g   %e",i,sqrt(nom));

    feti_projection (w,g,g1);
    
    // *******************************************************
    // *******************************************************
    //  vlozka s predpodminenim
    // *******************************************************
    // *******************************************************
    
    nullv (y,ndofcp);
    
    switch (prec){
    case noprecond:{
      for (j=0;j<ndofcp;j++){
	y[j]=w[j];
      }
      break;
    }
    case feti_lumped:{
      //scaling(w,w,ndofcp,out);
      for (j=0;j<ns;j++){
	nullv (wd,ndofmax);
	coarse_local (j,wd,w);
	//lscaling(dd,dd,j,out);
	nullv (pp,ndofmax);
	
	lumpedprec (j,wd,yd);
	
	//lscaling(pp,pp,j,out);
	local_coarse (j,yd,y);
      }
      //scaling(y,y,ndofcp,out);
      break;
    }
    default:{
      fprintf (stderr,"\n\n unknown type of preconditioner is required in function mpcg (file %s, line %d)\n",__FILE__,__LINE__);
    }
    }
    
    feti_projection (y,g,g1);
    
    // *******************************************************
    // *******************************************************
    //   konec vlozky s predpodminenim
    // *******************************************************
    // *******************************************************
    
    for (k=0;k<=i;k++){
      s=0.0;
      for (j=0;j<ndofcp;j++){
	s+=y[j]*ofp[k*ndofcp+j];
      }
      beta[k]=s/coef[k];
    }
    for (j=0;j<ndofcp;j++){
      s=0.0;
      for (k=0;k<=i;k++){
	s+=beta[k]*op[k*ndofcp+j];
      }
      p[j]=y[j]-s;
      op[(i+1)*ndofcp+j]=p[j];
    }
    
    
    if (sqrt(nom)<errcg){
      break;
    }
    
    
  }
  
  anicg=i;  aerrcg=nom;
  
  fprintf (out,"\n\n\n\n kontrola Lagrangeovych multiplikatoru \n");
  for (i=0;i<ndofcp;i++){
    //fprintf (out,"\n lagr. mult %4ld    %e",i,w[i]);
  }
  
}





/**
   function computes displacements from Lagrange multipliers
   function is used in FETI method
   
   @param w - array containing Lagrange multipliers
   @param d - array containing displacements
   @param f - array containing load forces
   @param rbm - array containing rigid body motions
   @param h - matrix H (columns are rigid body motions)
   @param h1 - matrix (H^T . H)^{-1}
   @param ndofcp - number of Lagrange multipliers
   @param hsize - number of columns of matrix H
   
   JK, 16.9.2007
*/
void seqfeti::lagrmultdispl (gtopology *top,gmatrix *gm,
			     double *w,double **d,double **ff,
			     double *h,double *h1,FILE *out)
{
  long i,j,k,l;
  double *g,*dd,*pp,*gamma,*u,*av;
  
  
  //av = new double [maxnrbm];
  av = new double [6];

  pp = new double [ndofmax];
  dd = new double [ndofmax];
  
  g = new double [ndofcp];
  
  nullv (g,ndofcp);
  for (j=0;j<ns;j++){
    nullv (pp,ndofmax);
    coarse_local (j,pp,w);
    
    subv (pp,ff[j],ndofdom[j]);
    
    smsky[j].ldl_feti_sky (d[j],pp,nrbmdom[j],se[j],zero);
    
    //  zacatek - streda 28.11.2007
    //  odecet skoku
    subv (d[j],jum[j],ndofdom[j]);
    //  konec - streda 28.11.2007
 
    local_coarse (j,d[j],g);
  }
  
  fprintf (out,"\n");
  for (i=0;i<ndofcp;i++){
    fprintf (out,"\n g  %20.15le",g[i]);
  }

  gamma = new double [hsize];
  u = new double [hsize];
  
  mtxv (h,g,u,ndofcp,hsize);
  mxv (h1,u,gamma,hsize,hsize);
  cmulv (-1.0,gamma,hsize);
  
  //for (j=0;j<ns;j++){
  //l=rbmadr[j];
  //for (i=0;i<nrbmdom[j];i++){
  //av[i] = gamma[l];  l++;
  //}
  //}
  
  //delete [] u;
  //delete [] gamma;
  //delete [] g;
  
  
  for (i=0;i<ns;i++){
    cmulv (-1.0,d[i],ndofdom[i]);
    
    nullv (av,6);
    l=rbmadr[i];
    for (j=0;j<nrbmdom[i];j++){
      av[j] = gamma[l];  l++;
    }

    fprintf (out,"\n");
    for (j=0;j<nrbmdom[i];j++){
      fprintf (out,"\n av %20.15le",av[j]);
    }

    for (j=0;j<ndofdom[i];j++){
      for (k=0;k<nrbmdom[i];k++){
	d[i][j]+=rbmdom[i][k*ndofdom[i]+j]*av[k];
      }
    }
  }
  

  //delete [] av;  delete [] pp;
}













/**
   function solves system of equations by the FETI method
   
   @param gm - %matrix of the system
   @param lhs - %vector of solution (left hand side)
   @param rhs - %vector of right hand side
   @param out - output stream
   
   JK, 16.9.2007
*/
void seqfeti::solve_system (gtopology *top,gmatrix *gm,
			    double *lhs,double *rhs,FILE *out)
{
  long i,j;
  double *h,*q,*hh,*ih,*ee,*lm;
  time_t t1,t2,t3,t4,t5,t6;
  
  //  determination of the ndofmax
  det_ndofmax ();

  //  allocation of vectors of right hand sides on subdomains
  double **ff;
  ff = new double* [ns];
  for (i=0;i<ns;i++){
    ff[i] = new double [ndofdom[i]];
    for (j=0;j<ndofdom[i];j++){
      ff[i][j]=0.0;
    }
  }
  
  
  //  allocation of vectors of right hand sides on subdomains
  for (i=0;i<ns;i++){
    globloc (rhs,ff[i],cndom[i],ndofdom[i]);
  }
  
  //  assembling of subdomain matrices
  subdomain_matrices (gm,out);
  
  t1 = time (NULL);
  
  //  computation of kernels / rigid body modes
  kernel (out);
  
  t2 = time (NULL);
  
  //  investigation of H matrix sizes
  hmatrixsize (out);
  
  hsize=rbmadr[ns];
  

  fprintf (stdout,"\n\n hsize   %ld \n",hsize);

  if (hsize==0){
    fprintf (stdout,"\n\n hsize = 0 \n");
    fprintf (stderr,"\n\n hsize = 0 \n");
    abort ();
  }
  
  h = new double [ndofcp*hsize];
  q = new double [hsize];
  lm = new double [ndofcp];
  
  fprintf (out,"\n hsize   %ld",hsize);
  fprintf (out,"\n ndofcp   %ld",ndofcp);
  
  for (i=0;i<=ns;i++){
    fprintf (out,"\n rbmadr %4ld   %4ld",i,rbmadr[i]);
  }
  
  //  assembling of H matrix
  hmatrix (h);
  
  fprintf (out,"\n\n\n kontrola matice H \n");
  for (i=0;i<ndofcp;i++){
    fprintf (out,"\n %4ld",i);
    for (j=0;j<hsize;j++){
      fprintf (out,"  %f",h[i*hsize+j]);
    }
  }
  
  //  assembling of q vector
  qvector (q,ff);
  
  
  fprintf (out,"\n\n\n kontrola vektoru q \n");
  for (j=0;j<hsize;j++){
    fprintf (out,"\n %4ld   %f",j,q[j]);
  }
  
  t3 = time (NULL);
  
  //  H^T . H and (H^T . H)^{-1.0} evaluation
  ih = new double [hsize*hsize];
  hh = new double [hsize*hsize];
  ee = new double [hsize*hsize];
  for (i=0;i<hsize;i++){
    for (j=0;j<hsize;j++){
      ee[i*hsize+j]=0.0;
    }
    ee[i*hsize+i]=1.0;
  }
  
  mtxm (h,h,hh,ndofcp,hsize,hsize);
  
  fprintf (out,"\n\n\n kontrola matice HH \n");
  for (i=0;i<hsize;i++){
    fprintf (out,"\n %4ld ",i);
    for (j=0;j<hsize;j++){
      fprintf (out," %lf",hh[i*hsize+j]);
    }
  }
  
  gemp (hh,ih,ee,hsize,hsize,zero,1);
  
  //mtm (h,h,hh,ndofcp,hsize,hsize);
  //mm (hh,ih,ee,hsize,hsize,hsize);
  
  
  fprintf (out,"\n\n\n kontrola matice HHinv \n");
  for (i=0;i<hsize;i++){
    fprintf (out,"\n %4ld ",i);
    for (j=0;j<hsize;j++){
      fprintf (out," %lf",ih[i*hsize+j]);
    }
  }
  
  delete [] ee;
  
  //  modified conjugate gradient method
  
  t4 = time (NULL);
  
  
  mpcg (top,gm,lm,ff,q,h,ih,out);
  //mprcg (top,gm,lm,ff,q,h,ih,out);
  
  t5 = time (NULL);
  
  double **dd;
  dd = new double* [ns];
  for (i=0;i<ns;i++){
    dd[i] = new double [ndofdom[i]];
    for (j=0;j<ndofdom[i];j++){
      dd[i][j]=0.0;
    }
  }
  
  lagrmultdispl (top,gm,lm,dd,ff,h,ih,out);
  
  
  for (i=0;i<ns;i++){
    for (j=0;j<ndofdom[i];j++){
      lhs[cndom[i][j]-1]=dd[i][j];
    }
  }
  
  fprintf (out,"\n\n kontrola");
  for (i=0;i<ns;i++){
    fprintf (out,"\n");
    for (j=0;j<ndofdom[i];j++){
      fprintf (out,"\n dd %4ld %4ld     %lf",i,j,dd[i][j]);
    }
  }
  
  
  t6 = time (NULL);
  
  
  fprintf (out,"\n\n evaluation of kernel             %ld",t2-t1);
  fprintf (out,"\n\n matrix H computation             %ld",t3-t2);
  fprintf (out,"\n\n decomposition of H               %ld",t4-t3);
  fprintf (out,"\n\n modified conjug. gradient m.     %ld",t5-t4);
  fprintf (out,"\n\n displacements computation        %ld",t6-t5);

  fprintf (stdout,"\n\n evaluation of kernel             %ld",t2-t1);
  fprintf (stdout,"\n\n matrix H computation             %ld",t3-t2);
  fprintf (stdout,"\n\n decomposition of H               %ld",t4-t3);
  fprintf (stdout,"\n\n modified conjug. gradient m.     %ld",t5-t4);
  fprintf (stdout,"\n\n displacements computation        %ld",t6-t5);

  
  
  
  //  vectors of right hand sides on subdomains
  for (i=0;i<ns;i++){
    delete [] ff[i];
  }
  delete [] ff;

}
