#ifndef SEQFETI_H
#define SEQFETI_H

#include "skyline.h"
//#include "gtopology.h"
//#include "gmatrix.h"
#include "seqselnodes.h"
#include "scr.h"
#include "densemat.h"

class gtopology;
class gmatrix;
//class symcomprow;

/**
   class feti deals with FETI method
   (Finite Element Tearing and Interconnecting method)
   this is single-processor implementation, it is different
   from feti1 in PARGEF which is multi-processor implementation
   
   
   
   JK, 21.8.2007
*/
class seqfeti
{
 public:
  seqfeti ();
  ~seqfeti (); 
  
  void read (XFILE *in);
  void print (FILE *out);

  void initiate (seqselnodes *selnodfeti,gtopology *top,FILE *out);
  void subdomain_ordering (gtopology *top,FILE *out);

  void det_ndofmax ();
 
  void assemble_subdom_unknowns (gtopology *top,FILE *out);
  void subdomain_matrices (gmatrix *gm,FILE *out);
  

  void local_coarse (long nd,double *lv,double *cv);
  void coarse_local (long nd,double *lv,double *cv);

  void kernel (FILE *out);


  void hmatrixsize (FILE *out);
  void hmatrix (double *h);


  void qvector (double *q,double **ff);
  
  void feti_projection (double *v,double *h,double *h1);


  void get_jumps (double *jump);


  void scaling(double *invect,double *outvector,long n,FILE *out);
  void lscaling(double *invect,double *outvect,long ndom,FILE *out);
  
  void lumpedprec (long nd,double *dd,double *pp);
  void dirichletprec (long nd,double *dd,double *pp);

  void mpcg (gtopology *top,gmatrix *gm,
	     double *w,double **ff,double *q,double *h,double *h1,FILE *out);
  void mprcg (gtopology *top,gmatrix *gm,
	      double *lambda,double **ff,double *e,double *g,double *g1,FILE *out);

  
  void lagrmultdispl (gtopology *top,gmatrix *gm,
		      double *w,double **d,double **ff,
		      double *h,double *h1,FILE *out);
  
  void solve_system (gtopology *top,gmatrix *gm,
		     double *lhs,double *rhs,FILE *out);
  
  
  ///  maximum number of iterations in conjugate gradient method
  long nicg;
  ///  number of performed iterations in conjugate gradient method
  long anicg;
  ///  required error
  double errcg;
  ///  attained error
  double aerrcg;
  ///  computer zero
  double zero;


  ///  array of numbers of unknowns (DOFs) contributing to the coarse problem
  ///  ncdofd contains nproc components
  ///  ncdofd[i]=j - the i-th subdomains contributes to coarse problem by j contributions
  long *ncdofd;
  
  ///  number of subdomains
  long ns;
  
  ///  type of preconditioner
  precondtype prec;

  long ndof;

  ///  number of DOFs (unknowns) in coarse problem
  long ndofcp;

  ///  estimated number of rigid body modes (estimated dimension of the kernel)
  long ense;
  
  ///  threshold for kernel detection
  double thresh;
  
  ///  maximum number of degrees of freedom on one subdomain
  long ndofmax;
  
  ///  size of matrix H
  long hsize;

  ///  numbers of nodes on subdomains
    //long *nnsd;
  ///  first node numbers on subdomains
    //long *fnnsd;
  
  


  
  
  ///  node-subdomain correspondence
  ///  nsid[i]=j - the i-th node belongs to the j-th subdomain
  long *nsid;
  
  
  ///  numbers of DOFs on subdomains
  ///  ndofdom[i]=j - the i-th subdomain contains j DOFs
  long *ndofdom;
  
  ///  list of DOFs on subdomains
  ///  cndom[i][j]=k - the j-th DOF on the i-th subdomain has number k
  long **cndom;
  
  
  ///  number of DOFs on subdomains used for preconditioning
  long *ndofprec;
  
  ///  code numbers for preconditioning
  ///  cnprec[i][j]=k - the j-th DOF on the i-th subdomain has global glued number k
  long **cnprec;
  
  ///  cpreccn[i][j]=k - the j-th DOF on the i-th subdomain has local number k
  long **cpreccn;

  double *wscalmat;
  long *nlwscalmat;
  double **lwscalmat;
  
  ///  array containing numbers of RBM on subdomains
  ///  nrbmdom[i]=j - the i-th subdomain contains j rigid body modes
  long *nrbmdom;

  ///  array containing addresses of first RBM in coarse matrix
  ///  rbmadr[i]=j - rigid body modes of the i-th subdomains start from the index j
  long *rbmadr;

  ///  list of linearly dependent equations
  ///  se[i][j]=k - the j-th base %vector of the kernel of the i-th subdomain creates the k-th column of the %matrix
  long **se;




  ///  array containing code numbers contributing to the coarse problem
  ///  extracted values from subdomains to the coarse problem
  ///  edofs[i][j]=k - the j-th components contributing to the coarse problem from the i-th subdomains has number k
  long **edofs;

  ///  array of coarse code numbers
  ///  it contains tnbn rows and ncdofd[i] columns
  ///  ccn[i][j]=k - the j-th contribution from the i-th subdomain goes to the k-th coarse unknown
  long **ccn;

  
  
  ///  subdomain matrices
  skyline *smsky;
  
  ///  rigid body modes / kernel
  double **rbmdom;
  
  ///  %matrix H
  double *h;
  
  ///  subdomain matrices for preconditioning
  ///  %matrix in the compressed row storage scheme
  symcomprow *smscr;
  ///  %matrix in the dense storage scheme (it contains Schur complements which are dense)
  densemat *smdm;

  ///  array containing jumps
  double **jum;
};

#endif
