#include "seqselnodes.h"
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

/**
   constructor
   
   @param nd - number of subdomains
   @param kk - number of nodes on subdomains
   @param jj - array containing list of selected nodes
               it contains nd,kk[i] components
   @param d - type of mesh description
   
   JK, 14.9.2007
*/
seqselnodes::seqselnodes(long nd,long *kk,long **jj,meshdescription d,FILE *out,long mespr)
{
  long i,j,k,ii;
  
  //  number of subdomains
  ns=nd;

  //  mesh description
  md = d;
  
  //  number of nodes on subdomains
  nnsd=kk;

  //  number of selected nodes
  nsndom = new long [ns];
  for (i=0;i<ns;i++){
    nsndom[i]=0;
    for (j=0;j<nnsd[i];j++){
      if (jj[i][j]>-1)
	nsndom[i]++;
    }
  }
  
  for (i=0;i<ns;i++){
    if (nsndom[i]==0){
      fprintf (stderr,"\n\n wrong number of selected nodes in constructor (file %s, line %d)\n",__FILE__,__LINE__);
      //abort();
    }
  }
  
  //  list of selected nodes - local numbers
  lsnl=new long* [ns];
  for (i=0;i<ns;i++){
    lsnl[i]=new long [nsndom[i]];
  }
  //  list of selected nodes - global numbers
  lsng=new long* [ns];
  for (i=0;i<ns;i++){
    lsng[i]=new long [nsndom[i]];
  }
  
  ii=0;
  for (i=0;i<ns;i++){
    k=0;
    for (j=0;j<nnsd[i];j++){
      if (jj[i][j]>-1){
	//lsnl[i][k]=j;
	lsnl[i][k]=ii;
	lsng[i][k]=jj[i][j];
	k++;
      }
      ii++;
    }
  }
  
  //if (mespr==0){
  fprintf (out,"\n\n\n list of local and global/group numbers of selected nodes (lsnl, lsng)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n number of selected nodes is nsndom %ld",nsndom[i]);
    for (j=0;j<nsndom[i];j++){
      fprintf (out,"\n local node number %6ld       global/group node number %6ld",lsnl[i][j],lsng[i][j]);
    }
  }
  //}
  
  
  //  total number of selected nodes
  tnsn=0;
  
  //  group node numbers
  gnn = NULL;
  
  //  node multiplicty
  nodmultip = NULL;
  
  // dof multiplicity
  dofmultip = NULL;
  
  //  numbers of DOFs
  ndofdom = NULL;
  
  //  number of DOFs on nodes
  ndofnmas = NULL;
  
  //  DOFs or indicators on master
  dofmas = NULL;
  

  //  array of numbers of DOFs for selected nodes
  ndofnsn=NULL;

  //  array of numbers of DOFs on subdomains 
  ndofdom = NULL;

  //  code numbers on master
  cndom=NULL;

  // ***********************************
  //  auxiliary arrays for FETI method
  // ***********************************

  //  list of joint nodes to selected nodes assumed as coarse nodes
  ljn=NULL;
  //  list of subdomain numbers which contain connected nodes to coarse nodes
  lsn=NULL;
  //  code numbers / indicators for FETI method
  doffeti=NULL;
  //  number of contributing nodes in the FETI method
  ncndom=NULL;
  //  list of code numbers which contribute to the coarse problem
  ldof = NULL;
  

  
}

/**
   destructor
   
   JK, 14.9.2007
*/
seqselnodes::~seqselnodes()
{
  long i,j;
  
  //  array cndom
  for (i=0;i<ns;i++){
    delete [] cndom[i];
  }
  delete [] cndom;
  
  //  array ldof
  for (i=0;i<ns;i++){
    delete [] ldof[i];
  }
  delete [] ldof;
  
  //  array ncndom
  delete [] ncndom;
  
  //  array ndofdom
  delete [] ndofdom;
  
  //  array doffeti
  for (i=0;i<tnsn;i++){
    for (j=0;j<nodmultip[i];j++){
      delete [] doffeti[i][j];
    }
    delete [] doffeti[i];
  }
  delete [] doffeti;
  
  //  array ndofnsn
  delete [] ndofnsn;
  
  //  array ljn
  for (i=0;i<tnsn;i++){
    delete [] ljn[i];
  }
  delete [] ljn;
  
  //  array lsn
  for (i=0;i<tnsn;i++){
    delete [] lsn[i];
  }
  delete [] lsn;
  
  //  array dofmas
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      delete [] dofmas[i][j];
    }
    delete [] dofmas[i];
  }
  delete [] dofmas;
  
  //  array ndofnmas
  for (i=0;i<ns;i++){
    delete [] ndofnmas[i];
  }
  delete [] ndofnmas;
  
  //  array ndofdom
  delete [] ndofdom;
  
  //  array nodmultip
  delete [] nodmultip;
  
  //  array gnn
  for (i=0;i<ns;i++){
    delete [] gnn[i];
  }
  delete [] gnn;
  
  //  array lsng
  for (i=0;i<ns;i++){
    delete [] lsng[i];
  }
  delete [] lsng;
  
  //  array lsnl
  for (i=0;i<ns;i++){
    delete [] lsnl[i];
  }
  delete [] lsnl;
  
  //  array nsndom
  delete [] nsndom;
}



/**
   function collects node numbers on master
   
   function assembles array gnn
   
   @param tnnp - total number of nodes in the problem
   @param out - output file
   
   14.9.2007, JK
*/
void seqselnodes::nodes_on_master (long tnnp,FILE *out)
{
  long i,j;
  long *nl;
  
  //  group node numbers on master
  if (gnn!=NULL){
    for (i=0;i<ns;i++){
      delete [] gnn[i];
    }
    delete [] gnn;
  }
  gnn = new long* [ns];
  for (i=0;i<ns;i++){
    gnn[i] = new long [nsndom[i]];
  }
  
  
  switch (md){
  case metis:
  case all_nodes:
  case neg_bound_nodes:{
    
    nl=new long [tnnp];
    for (i=0;i<tnnp;i++){
      nl[i]=-1;
    }
    
    for (i=0;i<ns;i++){
      for (j=0;j<nsndom[i];j++){
	nl[lsng[i][j]]++;
      }
    }
    
    //  total number of selected nodes
    //  generation of new node numbers
    tnsn=0;
    for (i=0;i<tnnp;i++){
      if (nl[i]>-1){
	nl[i]=tnsn;
	tnsn++;
      }
    }
    
    for (i=0;i<ns;i++){
      for (j=0;j<nsndom[i];j++){
	gnn[i][j]=nl[lsng[i][j]];
      }
    }
    
    delete [] nl;
    
    break;
  }
    
  case bound_nodes:{
    
    //  total number of selected nodes
    tnsn=0;
    for (i=0;i<ns;i++){
      for (j=0;j<nsndom[i];j++){
	gnn[i][j]=lsng[i][j];
	if (tnsn<gnn[i][j])
	  tnsn=gnn[i][j];
      }
    }
    tnsn++;
    
    break;
  }
  default:{
    fprintf (stderr,"\n\n wrong type of mesh description is required in functio");
    fprintf (stderr,"\n node_on_master (file %s, line %d)\n",__FILE__,__LINE__);
  }
  }
  
  fprintf (out,"\n\n total number of selected nodes is %ld",tnsn);
  fprintf (out,"\n\n group node numbers (gnn)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n\n subdomain %6ld",i+1);
    for (j=0;j<nsndom[i];j++){
      fprintf (out,"\n %6ld   %ld",j+1,gnn[i][j]);
    }
  }

}



/**
   function computes multiplicity of selected nodes
   
   function assembles array nodmultip
   
   JK, 14.9.2007
*/
void seqselnodes::node_multiplicity (FILE *out)
{
  long i,j;
  
  if (nodmultip!=NULL)
    delete [] nodmultip;
  nodmultip = new long [tnsn];
  for (i=0;i<tnsn;i++){
    nodmultip[i]=0;
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      nodmultip[gnn[i][j]]++;
    }
  }
  
  fprintf (out,"\n\n node multiplicity (nodmultip)\n");
  for (i=0;i<tnsn;i++){
    fprintf (out,"\n selected node %6ld    %ld",i+1,nodmultip[i]);
  }

}

/**
   function searches for all possible DOFs in selected nodes
   prescribed DOFs are included
   
   array ndofdom is assembled
   
   JK, 31.7.2007
*/
void seqselnodes::number_all_dofs (gtopology *top,FILE *out)
{
  long i,j,k;
  
  if (ndofdom!=NULL)
    delete [] ndofdom;
  ndofdom=new long [ns];
  
  for (i=0;i<ns;i++){
    ndofdom[i]=0;
    for (j=0;j<nsndom[i];j++){
      //  there is difference from parallel version
      //k=lsng[i][j];
      k=lsnl[i][j];
      ndofdom[i]+=top->give_ndofn (k);
    }
  }
  
  fprintf (out,"\n\n numbers of DOFs on subdomains (ndofdom)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n subdomain number %6ld   %ld",i+1,ndofdom[i]);
  }

}


/**
   function collects numbers of DOFs on selected nodes
   numbers are stored on master in the array ndofnmas
   
   array ndofnmas is assembled

   JK, 31.7.2007
*/
void seqselnodes::ndofn_on_master (gtopology *top,FILE *out)
{
  long i,j,k;

  //  group node numbers on master
  if (ndofnmas!=NULL){
    for (i=0;i<ns;i++){
      delete [] ndofnmas[i];
    }
    delete [] ndofnmas;
  }
  ndofnmas = new long* [ns];
  for (i=0;i<ns;i++){
    ndofnmas[i] = new long [nsndom[i]];
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      //k=lsng[i][j];
      k=lsnl[i][j];
      ndofnmas[i][j]=top->give_ndofn (k);
    }
  }
  
  fprintf (out,"\n\n numbers of DOFs on selected nodes (ndofnmas)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n subdomain number %6ld",i+1);
    for (j=0;j<nsndom[i];j++){
      fprintf (out,"\n selected node %6ld   %ld",j+1,ndofnmas[i][j]);
    }
  }

}


/**
   function assembles DOFs indicators on master
   
   array dofmas is assembled
   
   JK, 14.9.2007
*/
void seqselnodes::dof_indicators (gtopology *top,FILE *out)
{
  long i,j,k,l;
  
  dofmas = new long** [ns];
  for (i=0;i<ns;i++){
    dofmas[i] = new long* [nsndom[i]];
    for (j=0;j<nsndom[i];j++){
      dofmas[i][j] = new long [ndofnmas[i][j]];
    }
  }
    
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      for (k=0;k<ndofnmas[i][j];k++){
	//l=lsng[i][j];
	l=lsnl[i][j];
	dofmas[i][j][k]=top->give_dof (l,k);
      }
    }
  }

  fprintf (out,"\n\n DOFs indicators on master (dofmas)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n subdomain number %6ld",i+1);
    for (j=0;j<nsndom[i];j++){
      fprintf (out,"\n selected node %6ld    ",j+1);
      for (k=0;k<ndofnmas[i][j];k++){
	fprintf (out,"  %ld",dofmas[i][j][k]);
      }
    }
  }

}



/**
   function assembles lists of connected nodes to coarse nodes
   and numbers of subdomains which contain connected nodes
   
   arrays ljn and lsn are assembled
   
   JK, 14.9.2007
*/
void seqselnodes::group_local_nodes (FILE *out)
{
  long i,j,k,m,n,min;
  
  ljn = new long* [tnsn];
  lsn = new long* [tnsn];
  for (i=0;i<tnsn;i++){
    ljn[i] = new long [nodmultip[i]];
    lsn[i] = new long [nodmultip[i]];
    nodmultip[i]=0;
  }
  
  //  list of local node numbers of connected nodes
  //  list of numbers of subdomains which contains connected nodes
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      k=gnn[i][j];
      ljn[k][nodmultip[k]]=j;
      lsn[k][nodmultip[k]]=i;
      nodmultip[k]++;
    }
  }
  
  
  //  node sorting
  for (i=0;i<tnsn;i++){
    for (j=0;j<nodmultip[i];j++){
      min=ns;
      for (k=j;k<nodmultip[i];k++){
	if (lsn[i][k]<min){
	  min=lsn[i][k];
	  m=k;
	}
      }
      n=lsn[i][j];
      lsn[i][j]=lsn[i][m];
      lsn[i][m]=n;
      
      n=ljn[i][j];
      ljn[i][j]=ljn[i][m];
      ljn[i][m]=n;
    }
  }
  
  fprintf (out,"\n\n list of local node numbers of connected nodes (ljn)\n");
  fprintf (out,"\n\n list of numbers of subdomains which contains connected nodes (lsn)\n");
  for (i=0;i<tnsn;i++){
    fprintf (out,"\n coarse node %6ld   multip %6ld   ",i+1,nodmultip[i]);
    for (j=0;j<nodmultip[i];j++){
      fprintf (out,"    %ld %ld",ljn[i][j],lsn[i][j]);
    }
  }

}

/**
   function assembles indicators of code numbers and then generates
   code numbers
   
   JK, 14.9.2007
*/
void seqselnodes::dof_feti (FILE *out)
{
  long i,j,k,l,m;
  
  //  determination of numbers of DOFs at selected nodes
  if (ndofnsn!=NULL)
    delete [] ndofnsn;
  ndofnsn = new long [tnsn];
  for (i=0;i<tnsn;i++){
    ndofnsn[i]=0;
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      k=ndofnmas[i][j];
      l=gnn[i][j];
      if (ndofnsn[l]>0){
	if (ndofnsn[l]!=k){
	  fprintf (stderr,"\n\n incompatible number of DOFs at selected node number %ld (file %s, line %d)\n",l,__FILE__,__LINE__);
	}
      }
      else{
	ndofnsn[l]=k;
      }
    }
  }
  
  //fprintf (out,"\n\n kontrola ndofnsn");
  //for (i=0;i<tnsn;i++){
  //fprintf (out,"\n %ld   %ld",i+1,ndofnsn[i]);
  //}
  
  
  //  array allocation
  doffeti = new long** [tnsn];
  for (i=0;i<tnsn;i++){
    doffeti[i] = new long* [nodmultip[i]];
    for (j=0;j<nodmultip[i];j++){
      doffeti[i][j] = new long [ndofnsn[i]];
    }
  }
  
  //  assembling of code numbers indicators
  for (i=0;i<tnsn;i++){
    for (j=0;j<nodmultip[i];j++){
      l=ljn[i][j];
      m=lsn[i][j];
      for (k=0;k<ndofnsn[i];k++){
	doffeti[i][j][k]=dofmas[m][l][k];
      }
    }
  }
  
  //  code numbers generation
  tndof=1;
  for (i=0;i<tnsn;i++){
    for (j=0;j<nodmultip[i]-1;j++){
      for (k=0;k<ndofnsn[i];k++){
	if (doffeti[i][j][k]>0){
	  doffeti[i][j][k]=tndof;
	  tndof++;
	}
      }
    }
  }
  tndof--;
  
  fprintf (out,"\n\n code numbers for FETI method (doffeti)\n");
  for (i=0;i<tnsn;i++){
    fprintf (out,"\n coarse node %6ld",i+1);
    for (j=0;j<nodmultip[i]-1;j++){
      fprintf (out,"  con.n. %6ld  sub.n. %6ld   ",ljn[i][j],lsn[i][j]);
      for (k=0;k<ndofnsn[i];k++){
	fprintf (out," %ld",doffeti[i][j][k]);
      }
    }
  }
  
}


void seqselnodes::dof_multiplicity (FILE *out)
{
  long i,j,k,n;
  long ndofn;
  long nm;
  
  if(dofmultip != NULL){
    delete []dofmultip;
  }
  
  
  dofmultip = new long[tndof];
  n = 0;
  
  for(i = 0; i < tnsn; i++){
    nm = nodmultip[i];
    for(j = 0; j < nm-1; j++){
      ndofn = ndofnsn[i];
      for(k = 0; k < ndofn; k++){
	if(doffeti[i][j][k] > 0){
	  dofmultip[n] = nm;
	  n++;
	}
      }
    }
  }
  /*
  fprintf (out,"\n\n\n Multiplicity of coarse DOF n = %ld\n\n",n); 
  for(i = 0; i < tndof; i++){
    fprintf (out,"coarse DOF %ld has multiplicity %ld\n",i+1,dofmultip[i]); 
  }
  */
  /*
  ldofmultip = new long*[ns];
  for(i = 0; i < ns; i++){
    ldofmultip[i] = new long[ndofdom[i]];
    for(j = 0; j < ndofdom[i]; j++){
      ldofmultip[i][j] =  dofmultip[cndom[i][j]-1];
    }
  }
  */
  /*
  fprintf (out,"\n\n\n local multiplicity of coarse DOF\n\n"); 
  for(i = 0; i < ns; i++){
    fprintf (out,"domain %ld\n",i+1);
    for(j = 0; j < ndofdom[i]; j++){
      fprintf (out,"%ld %ld %ld\n",j+1,cndom[i][j],ldofmultip[i][j]); 
    }
  }
  */

}

/**
   function determines number of contributions to coarse problem
   from particular subdomains
   
   maxndof is rewritten
   ndofdom is rewritten
   
   array ncn is assembled
   
   JK, 14.9.2007
*/
void seqselnodes::number_contrib (FILE *out)
{
  long i,j,k,l,m,ii;
  
  if (ndofdom!=NULL)
    delete [] ndofdom;
  ndofdom = new long [ns];
  for (i=0;i<ns;i++){
    ndofdom[i]=0;
  }
  
  if (ncndom!=NULL)
    delete [] ncndom;
  ncndom = new long [ns];
  for (i=0;i<ns;i++){
    ncndom[i]=0;
  }
  
  //  number of contributing DOFs from subdomains to coarse problem
  for (i=0;i<ns;i++){
    for (j=0;j<nsndom[i];j++){
      ii=gnn[i][j];
      for (k=0;k<nodmultip[ii];k++){
	if (lsn[ii][k]==i){
	  if (k==0){
	    for (l=0;l<nodmultip[ii]-1;l++){
	      ncndom[i]++;
	      for (m=0;m<ndofnsn[ii];m++){
		if (doffeti[ii][l][m]>0)
		  ndofdom[i]++;
	      }
	    }
	  }
	  else{
	    ncndom[i]++;
	    for (m=0;m<ndofnsn[ii];m++){
	      if (doffeti[ii][k-1][m]>0)
		ndofdom[i]++;
	    }
	  }
	  break;
	}
      }
    }
  }
    
  fprintf (out,"\n\n number of contributing DOFs to coarse problem (ndofdom)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n subdomain number %6ld   %ld",i+1,ndofdom[i]);
  }
  fprintf (out,"\n\n number of nodes which contribute to coarse FETI problem (ncn)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n domain %ld   %ld",i,ncndom[i]);
  }
  
}




/**
   function assmebles nodes contributing to the coarse proble in the FETI method
      
   JK, 14.9.2007
*/
void seqselnodes::contrib_dofs (gtopology *top,FILE *out)
{
  long i,j,k,l,m,ii,jj,kk,ndofn,min;
  long *buff;
  
  //ndofdom = new long [ns];
  ldof = new long* [ns];

  //  array of local numbers of nodes contributing to the coarse problem
  for (i=0;i<ns;i++){
    
    buff = new long [ncndom[i]];
    for (j=0;j<ncndom[i];j++){
      buff[j]=0;
    }
    
    m=0;
    for (j=0;j<tnsn;j++){
      for (k=0;k<nodmultip[j];k++){
	if (lsn[j][k]==i){
	  if (k==0){
	    for (l=0;l<nodmultip[j]-1;l++){
	      buff[m]=ljn[j][k];
	      buff[m]=0-buff[m]-1;
	      m++;
	    }
	  }
	  else{
	    buff[m]=ljn[j][k];
	    m++;
	  }
	  break;
	}
      }
    }
    
    
    for (ii=0;ii<ncndom[i];ii++){
      if (buff[ii]<0)
	min=0-buff[ii]-1;
      else
	min=buff[ii];
      
      l=ii;
      for (j=ii+1;j<ncndom[i];j++){
	if (buff[j]<0)
	  k=0-buff[j]-1;
	else
	  k=buff[j];
	
	if (min>k){
	  l=j;
	  min=k;
	}
      }
      
      
      k=buff[ii];
      buff[ii]=buff[l];
      buff[l]=k;
    }
    
    fprintf (out,"\n\n kontrola bufferu na vsech uzlech");
    for (j=0;j<ncndom[i];j++){
      fprintf (out,"\n %ld   %ld",j+1,buff[j]);
    }
    //fprintf (out,"\n konec kontroly\n");
    //fprintf (out,"\n\n kontrola lsnl na vsech uzlech");
    //for (i=0;i<nsn;i++){
    //fprintf (out,"\n %ld   %ld",i+1,lsnl[i]);
    //}
    //fprintf (out,"\n konec kontroly\n");
    
    
    
    //  number of contributions to the coarse problem
    //  number of DOFs which contribute to the coarse problem
    //ndofdom[i]=0;
    //for (ii=0;ii<ncndom[i];ii++){
    //j=buff[ii];
    //if (j<0){
    //j=0-j-1;
    //}
      //l=lsng[i][j];
      //l=lsnl[i][j];
      //ndofn=top->give_ndofn (l);
      //for (k=0;k<ndofn;k++){
    //if (top->give_dof (l,k)>0)
    //ndofdom[i]++;
    //}
    //}
    
    //  list of code numbers which are extracted from domain in the FETI method
    //  some code numbers are positive and some of them are negative
    //  it depends on number of subdomain
    //  these code numbers are used on subdomains, master processor contains corresponding array with coarse code numbers
    
    ldof[i] = new long [ndofdom[i]];
    
    m=0;
    for (kk=0;kk<ncndom[i];kk++){
      ii=buff[kk];
      if (ii<0){
	ii=0-ii-1;
	//jj=lsng[i][ii];
	jj=lsnl[i][ii];
	ndofn=top->give_ndofn (jj);
	for (j=0;j<ndofn;j++){
	  l=top->give_dof (jj,j);
	  if (l>0){
	    ldof[i][m]=0-l;
	    m++;
	  }
	}
      }
      else{
	//jj=lsng[i][ii];
	jj=lsnl[i][ii];
	ndofn=top->give_ndofn (jj);
	for (j=0;j<ndofn;j++){
	  l=top->give_dof (jj,j);
	  if (l>0){
	    ldof[i][m]=l;
	    m++;
	  }
	}
      }
    }
    
    delete [] buff;
  }

  
  cndom = new long* [ns];
  for (i=0;i<ns;i++){
    cndom[i] = new long [ndofdom[i]];
  }
  
  for (i=0;i<ns;i++){
    l=0;
    for (j=0;j<nsndom[i];j++){
      m=gnn[i][j];
      for (k=0;k<nodmultip[m];k++){
	if (lsn[m][k]==i){
	  if (k==0){
	    for (ii=0;ii<nodmultip[m]-1;ii++){
	      for (jj=0;jj<ndofnsn[m];jj++){
		if (doffeti[m][ii][jj]>0){
		  cndom[i][l]=doffeti[m][ii][jj];
		  l++;
		}
	      }
	    }
	  }
	  else{
	    for (jj=0;jj<ndofnsn[m];jj++){
	      if (doffeti[m][k-1][jj]>0){
		cndom[i][l]=doffeti[m][k-1][jj];
		l++;
	      }
	    }
	  }
	  break;
	}
      }
    }
  }
  
  fprintf (out,"\n array ldof");
  for (i=0;i<ns;i++){
    fprintf (out,"\n\n %ld   number of DOFs which contribute to coarse FETI problem %ld",i,ndofdom[i]);
    for (j=0;j<ndofdom[i];j++){
      fprintf (out,"\n %6ld   %ld",j+1,ldof[i][j]);
    }
  }
  
  fprintf (out,"\n\n code numbers on master (cndom)\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n subdomain number %6ld (number of DOFs %ld)",i+1,ndofdom[i]);
    for (j=0;j<ndofdom[i];j++){
      fprintf (out,"\n %6ld   %ld",j+1,cndom[i][j]);
    }
  }

}

/**
   function prepares all data for the FETI method
   
   @param gt - pointer to the general topology
   @param out - output file
   
   JK, 18.11.2007
*/
void seqselnodes::prepare_feti (gtopology *gt,FILE *out)
{
  nodes_on_master (gt->nn,out);
  node_multiplicity (out);
  number_all_dofs (gt,out);
  ndofn_on_master (gt,out);
  dof_indicators (gt,out);
  group_local_nodes (out);
  dof_feti (out);
  //dof_multiplicity (out);
  number_contrib (out);
  contrib_dofs (gt,out);
  dof_multiplicity (out);
}
