#ifndef SEQSELNODES_H
#define SEQSELNODES_H

#include <stdio.h>
#include "galias.h"
#include "gtopology.h"

/**
   class deals with selected nodes on subdomains
   the class is sequential version of the class selnodes
   nodes could be selected for Schur complement method,
   FETI method, etc.
   
   JK, 14.9.2007
*/
class seqselnodes
{
 public:
  seqselnodes (long nd,long *kk,long **jj,meshdescription d,FILE *out,long mespr);
  ~seqselnodes ();

  void nodes_on_master (long tnnp,FILE *out);
  void node_multiplicity (FILE *out);
  void number_all_dofs (gtopology *top,FILE *out);
  void ndofn_on_master (gtopology *top,FILE *out);
  void dof_indicators (gtopology *top,FILE *out);
  void group_local_nodes (FILE *out);
  void dof_feti (FILE *out);
  void dof_multiplicity (FILE *out);
  void number_contrib (FILE *out);
  void contrib_dofs (gtopology *top,FILE *out);

  void prepare_feti (gtopology *gt,FILE *out);

  
  ///  type of mesh description
  meshdescription md;

  ///  number of subdomains
  long ns;
    
  ///  total number of selected nodes
  long tnsn;

  ///  total number of DOFs
  long tndof;

  ///  numbers of nodes on subdomains
  long *nnsd;

  ///  numbers of selected nodes
  ///  nsndom[i]=j - there are j selected nodes on the i-th subdomain
  long *nsndom;

  ///  list of selected nodes - local numbers
  ///  lsnl[i][j]=k - the j-th selected node on the i-th subdomain has local number k
  ///  lsnl contains ns,nsn[i] components
  long **lsnl;

  ///  list of selected nodes - global numbers
  ///  lsng[i][j]=k - the j-th selected node on the j-th subdomain has global/coarse number k
  ///  lsng contains ns, nsn[i] components
  long **lsng;

  ///  group node numbers (see partop.h)
  ///  gnn[i][j]=k - the j-th selected node on the i-th subdomain has group number k
  long **gnn;

  ///  node multiplicity
  ///  nodmultip[i]=j - the i-th selected node is shared by j subdomains
  ///  it contains tnsn components
  long *nodmultip;
  
  long *dofmultip;
  long **ldofmultip;
  
  ///  array of numbers of DOFs on subdomains 
  ///  it contains prescribed values before code number generation
  ///  after code number generation, only unknown (unconstrained DOFs) are taken into account
  ///  array is rewritten in the function schur_ordering
  ///  ndofdom[i]=j - the i-th subdomain contains j DOFs
  long *ndofdom;

  ///  array of DOFs at selected nodes
  ///  ndofnmas[i][j]=k - the j-th node on the i-th subdomain contains k DOFs
  long **ndofnmas;

  ///  array of DOFs or indicators on master
  ///  dofmas[i][j][k]=l - the k-th DOF at the j-th selected node on the i-th subdomain has value l
  long ***dofmas;

  ///  array of numbers of DOFs for selected nodes
  ///  ndofnsn[i]=j - the i-th selected node (in group ordering) has j DOFs
  ///  it contains tnsn components
  long *ndofnsn;

  ///  code numbers on master
  ///  cndom[i][j]=k - the j-th DOF on the i-th subdomain has group code number k
  long **cndom;


  //  FETI ORDERING

  ///  list of joint nodes to selected nodes assumed as coarse nodes
  ///  ljn[i][j]=k - the j-th node connected to the i-th coarse node has local number k
  ///  ljn contains tnsn rows and nodmultip columns
  long **ljn;
  
  ///  list of subdomain numbers which contain connected nodes to coarse nodes
  ///  lsn[i][j]=k - the j-th node connected to the i-th coarse node belongs to the k-th subdomain
  ///  lsn contains tnsn rows and nodmultip columns
  long **lsn;

  ///  code numbers / indicators for FETI method
  ///  doffeti[i][j][k]=l - the k-th DOF on the j-th connected node to the i-th coarse node has code number / indicator l
  long ***doffeti;
  
  ///  number of contributing nodes in the FETI method
  ///  ncndom[i]=j - the i-th subdomain contributes to the coarse problem by j nodes
  long *ncndom;

  ///  list of code numbers which contribute to the coarse problem
  ///  ldof[i][j]=k - the j-th DOF on the i-th subdomain has local code number k
  long **ldof;

};

#endif

