#include "seqtop.h"
#include <string.h>
#include <math.h>
#include <time.h>
#include "galias.h"
#include "iotools.h"
#include "gtopology.h"

/**
   constructor
   
   @param nd - number of subdomains
   @param meshd - mesh description
   
   JK, 11.9.2007
*/
seqtop::seqtop (long nd,meshdescription meshd)
{
  //  number of subdomains
  ns=nd;
  //  mesh description
  md = meshd;


  //  total number of nodes in the whole problem
  tnnp=0;
  //  total number of interface (boundary) nodes
  tnbn=0;
  
  
  //  array containing number of nodes on subdomains
  nnsd=NULL;

  //  array containing list of numbers of boundary nodes on subdomains
  nbnd=NULL;

  //  array containing list of numbers of internal nodes on subdomains
  nind=NULL;
  
  //  array containing numbers of boundary nodes
  lnbn=NULL;
  
  //  array containing numbers of internal nodes
  lnin=NULL;
  
  //  array containing list of global numbers of interface nodes
  lgnbn=NULL;
  
  
  //  array containing number of subdomains which each boundary/interface node belongs to
  multip=NULL;

  //  array containing number of subdomains which share the nodes
  nodmultip=NULL;



  //  local to global correspondence
  ltg=NULL;
}

/**
   destructor
   
   JK, 11.9.2007
*/
seqtop::~seqtop()
{
 
}


/**
   function reads array containing numbers of nodes on subdomains
   
   @param in - input file
   
   JK, 11.9.2007
*/
void seqtop::read_nnsd (XFILE *in)
{
  long i;

  if (nnsd!=NULL){
    delete [] nnsd;
  }
  nnsd = new long [ns];
  
  for (i=0;i<ns;i++){
    xfscanf (in,"%ld",nnsd+i);
  }
  
}


/**
   function reads array ltg (local to global map)
   
   @param in - input file
   
   11.9.2007, JK
*/
void seqtop::read_ltg (XFILE *in)
{
  long i,j;
  long nn;

  if (ltg!=NULL){
    for (i=0;i<ns;i++){
      delete [] ltg[i];
    }
    delete [] ltg;
  }
  ltg = new long* [ns];
  for (i=0;i<ns;i++){
    ltg[i] = new long [nnsd[i]];
  }
  
  
  if (md==metis){
    nn=0;
    for (i=0;i<ns;i++){
      nn+=nnsd[i];
      nnsd[i]=0;
    }
    
    for (i=0;i<nn;i++){
      xfscanf (in,"%ld",&j);
      ltg[j][nnsd[j]]=i;
      nnsd[j]++;
    }
  }
  else{
    for (i=0;i<ns;i++){
      for (j=0;j<nnsd[i];j++){
	xfscanf (in,"%ld",&ltg[i][j]);
	ltg[i][j]--;
      }
    }
  }
}


/**
   function reads array ltg (local to global map)
   
   @param in - input file
   
   11.9.2007, JK
*/
void seqtop::read_loadbal_ltg (XFILE *in)
{
  long i,j;
  long nn;

  if (ltg!=NULL){
    for (i=0;i<ns;i++){
      delete [] ltg[i];
    }
    delete [] ltg;
  }
  ltg = new long* [ns];
  for (i=0;i<ns;i++){
    ltg[i] = new long [nnsd[i]];
  }
  
  nn=0;
  for (i=0;i<ns;i++){
    nn+=nnsd[i];
  }
  
  for (i=0;i<ns;i++){
    for (j=0;j<nnsd[i];j++){
      xfscanf (in,"%ld",&ltg[i][j]);
      ltg[i][j]--;
    }
  }
}

/**
   function computes node multiplicity
   function is used only for mesh description = all_nodes or neg_bound_nodes
   
   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);

   
   function establishes:
   maxnbn - maximum number of boundary nodes
   
   nbn - number of boundary nodes on each subdomain
   
   tnbn - total number of boundary nodes
   
   nbnd - array containing numbers of boundary nodes on subdomains
   
   tnnp - total number of nodes in problem
   
   multip - array of boundary node multiplicity (on the master)
   
   nodmultip - array of node multiplicity (on all processors)
   
   
   @param top - pointer to subdomain topology
   @param domproc - array containing domain-processor correspondence
   
   JK, 11.9.2007
*/
void seqtop::compute_multiplicity (FILE *out)
{
  long i,j,k;
  
  //  list of number of boundary nodes on subdomains
  if (nbnd != NULL)
    delete [] nbnd;
  nbnd = new long [ns];
  
  
  switch (md){
  case bound_nodes:{
    
    tnbn=0;
    for (i=0;i<ns;i++){
      nbnd[i]=0;
      for (j=0;j<nnsd[i];j++){
	if (tnbn<ltg[i][j])
	  tnbn=ltg[i][j];
	if (ltg[i][j]>-1)
	  nbnd[i]++;
      }
    }
    tnbn++;
    
    if (multip!=NULL){
      delete [] multip;
    }
    multip = new long [tnbn];
    for(i=0;i<tnbn;i++){
      multip[i]=0;
    }
    
    for (i=0;i<ns;i++){
      for (j=0;j<nnsd[i];j++){
	k=ltg[i][j];
	if (k>-1)
	  multip[k]++;
      }
    }
    
    if (nodmultip!=NULL){
      for (i=0;i<ns;i++){
	delete [] nodmultip[i];
      }
      delete [] nodmultip;
    }
    nodmultip = new long* [ns];
    for (i=0;i<ns;i++){
      nodmultip[i] = new long [nnsd[i]];
    }
    
    for (i=0;i<ns;i++){
      for (j=0;j<nnsd[i];j++){
	k=ltg[i][j];
	if (k>-1){
	  nodmultip[i][j]=multip[k];
	}
	else{
	  nodmultip[i][j]=1;
	}
      }
    }
    
       
    break;
  }



  case metis:
  case all_nodes:
  case neg_bound_nodes:{
    
    tnnp=0;
    for (i=0;i<ns;i++){
      for (j=0;j<nnsd[i];j++){
	if (tnnp<ltg[i][j])
	  tnnp=ltg[i][j];
      }
    }
    //  must be increased due to indices from 0 instead of 1
    tnnp++;
    
    if (multip != NULL)
      delete [] multip;
    multip = new long [tnnp];
    for (i=0;i<tnnp;i++){
      multip[i]=0;
    }
    
    //  computation of node incidences
    for (i=0;i<ns;i++){
      for (j=0;j<nnsd[i];j++){
	multip[ltg[i][j]]++;
      }
    }
    
    tnbn=0;
    for (i=0;i<tnnp;i++){
      if (multip[i]>1)
	tnbn++;
    }
    
    
    //  nodal multiplicity
    if (nodmultip!=NULL){
      for (i=0;i<ns;i++){
	delete [] nodmultip[i];
      }
      delete [] nodmultip;
    }
    nodmultip = new long* [ns];
    for (i=0;i<ns;i++){
      nodmultip[i] = new long [nnsd[i]];
    }

    
    for (i=0;i<ns;i++){
      nbnd[i]=0;
      for (j=0;j<nnsd[i];j++){
	k=multip[ltg[i][j]];
	nodmultip[i][j]=k;
	if (k>1)
	  nbnd[i]++;
      }
    }
    
    break;
  }
  default:{
    print_err("unknown type of mesh description is required", __FILE__, __LINE__, __func__);
  }
  }
  
  
  fprintf(out,"\n\n\n Multiplicity of nodes on subdomain\n\n");  
  for (i=0;i<ns;i++){
    for (j=0;j<nnsd[i];j++){
      fprintf(out,"%4ld  %6ld   %ld\n",i,j,nodmultip[i][j]);  
    }
  }
  
  fprintf(out,"\n\n\n total number of boundary nodes in whole problem is %ld\n\n",tnbn);
  
  fprintf (out,"\n\n\n numbers of boundary nodes on each subdomain \n\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n %ld",nbnd[i]);
  }
  
}

/**
   function selects boundary nodes
   
   the following functions have to be called before this function:
   void initiation (gtopology *top,long *ltg);
   void numbers_of_all_nodes_on_subdomains (gtopology *top,long *domproc,FILE *out);
   void compute_multiplicity (gtopology *top,long *domproc,FILE *out);

   
   function establishes:
   
   lnbn - local numbers of boundary nodes
   
   @param out - output stream
   
   JK, 14.9.2007
*/
void seqtop::find_boundary_nodes (FILE *out)
{
  long i,j,k,*buff;
  
  //  list of coarse numbers of boundary nodes
  if (lgnbn!=NULL){
    for (i=0;i<ns;i++){
      delete [] lgnbn[i];
    }
    delete [] lgnbn;
  }
  lgnbn = new long* [ns];
  for (i=0;i<ns;i++){
    lgnbn[i] = new long [nbnd[i]];
  }
  
  
  switch (md){
  case bound_nodes:{
    for (i=0;i<ns;i++){
      k=0;
      for (j=0;j<nnsd[i];j++){
	if (ltg[i][j]>-1){
	  lgnbn[i][k]=ltg[i][j];
	  k++;
	}
      }
    }
    break;
  }
  case metis:
  case all_nodes:
  case neg_bound_nodes:{
    
    for (i=0;i<ns;i++){
      k=0;
      for (j=0;j<nnsd[i];j++){
	if (multip[ltg[i][j]]>1){
	  lgnbn[i][k]=ltg[i][j];
	  k++;
	}
      }
    }
    
    break;
  }
  default:{
    print_err("unknown type of mesh description is required", __FILE__, __LINE__, __func__);
  }
  }


  // *******************************
  //  indication of boundary nodes 
  // *******************************
  //  local numbers of boundary nodes
  if (lnbn != NULL){
    for (i=0;i<ns;i++){
      delete [] lnbn[i];
    }
    delete [] lnbn;
  }
  lnbn = new long* [ns];
  for (i=0;i<ns;i++){
    lnbn[i] = new long [nbnd[i]];
  }
  
  switch (md){
  case metis:
  case all_nodes:
  case neg_bound_nodes:{
    for (i=0;i<ns;i++){
      k=0;
      for (j=0;j<nnsd[i];j++){
	if (nodmultip[i][j]>1){
	  lnbn[i][k]=j;
	  k++;
	}
      }
    }
    break;
  }
    
  case bound_nodes:{
    for (i=0;i<ns;i++){
      k=0;
      for (j=0;j<nnsd[i];j++){
	if (ltg[i][j]>-1){
	  lnbn[i][k]=j;
	  k++;
	}
      }
    }
    break;
  }
  default:{
    print_err("unknown type of mesh description is required", __FILE__, __LINE__, __func__);
  }
  }
  
  if (nind!=NULL){
    delete [] nind;
  }
  nind = new long [ns];

  //  number of internal nodes
  for (i=0;i<ns;i++){
    nind[i] = nnsd[i]-nbnd[i];
  }
  
  //  array containing local numbers of internal nodes
  if (lnin != NULL){
    for (i=0;i<ns;i++){
      delete [] lnin[i];
    }
    delete [] lnin;
  }
  lnin = new long* [ns];
  for (i=0;i<ns;i++){
    lnin[i] = new long [nind[i]];
  }
  
  for (i=0;i<ns;i++){
    buff = new long [nnsd[i]];
    for (j=0;j<nnsd[i];j++){
      buff[j]=0;
    }
    for (j=0;j<nbnd[i];j++){
      k=lnbn[i][j];
      buff[k]=1;
    }
    
    k=0;
    for (j=0;j<nnsd[i];j++){
      if (buff[j]==0){
	lnin[i][k]=j;
	k++;
      }
    }
    delete [] buff;
  }
  
  fprintf (out,"\n\n\n numbers of interface nodes on subdomains \n\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n %ld",nbnd[i]);
  }
  
  fprintf (out,"\n\n\n local numbers of interface nodes on subdomain");
  for (i=0;i<ns;i++){
    fprintf (out,"\n\n domain %ld",i);
    for (j=0;j<nbnd[i];j++){
      fprintf (out,"\n %6ld   %6ld",j,lnbn[i][j]+1);
    }
  }
  
  fprintf (out,"\n\n\n numbers of internal nodes on subdomains \n\n");
  for (i=0;i<ns;i++){
    fprintf (out,"\n %ld",nind[i]);
  }
  
  fprintf (out,"\n\n\n local numbers of internal nodes on subdomain");
  for (i=0;i<ns;i++){
    for (j=0;j<nind[i];j++){
      fprintf (out,"\n %6ld   %6ld",j,lnin[i][j]+1);
    }
  }
}


/**
   function rewrites array ltg
   
   JK, 14.9.2007
*/
void seqtop::rewrite_ltg ()
{
  long i,j;
  
  for (i=0;i<ns;i++){
    for (j=0;j<nnsd[i];j++){
      ltg[i][j]=-1;
    }
  }
  for (i=0;i<ns;i++){
    for (j=0;j<nbnd[i];j++){
      ltg[i][lnbn[i][j]]=lgnbn[i][j];
    }
  }
}

/**
   function renumbers code numbers
   
   JK, 9.10.2007
*/
void seqtop::codnum_renumb (gtopology *top)
{
  long i,j,k,ndofn,dof,nid;
  long *red;
  red = new long [ns+1];
  
  red[0]=0;
  nid=0;
  for (i=0;i<ns;i++){
    red[i+1]=0;
    for (j=0;j<nnsd[i];j++){
      ndofn = top->give_ndofn (nid);
      for (k=0;k<ndofn;k++){
	dof = top->give_dof (nid,k);
	if (dof>red[i+1]){
	  red[i+1]=dof;
	}
      }
      nid++;
    }
  }
  
  nid=0;
  for (i=0;i<ns;i++){
    for (j=0;j<nnsd[i];j++){
      ndofn = top->give_ndofn (nid);
      for (k=0;k<ndofn;k++){
	dof = top->give_dof (nid,k);
	if (dof>0){
	  dof-=red[i];
	  top->save_dof (nid,k,dof);
	}
      }
      nid++;
    }
  }
  
  delete [] red;
}

/**
   function creates (assembles) array ltg
   (local to global array)
   
   @param gt - general topology
   @param out - output file
   
   JK, 13.10.2007
*/
/*
void seqtop::create_ltg (gtopology *gt,FILE *out)
{
  long i,j,k,nn,nedg,nin;
  long *aux,*v;

  //  number of nodes in gtopology
  //  some nodes are computed several times
  //  nn is not equal to the number of nodes in the original undecomposed mesh
  nn = gt->nn;

  aux = new long [nn];
  for (i=0;i<nn;i++){
    aux[i]=nn;
  }
  
  //  number of interface edges
  nedg = gt->nged;
  
  //  auxiliary array assembling
  //  aux[i]=j - the i-th node has label j
  //  if j=nn - noninterface node
  //  if j!=nn - interface node, j is equal to minimum number of node connected to this interface node
  //
  for (i=0;i<nedg;i++){
    j=gt->gedges[i].nlist[0];
    k=gt->gedges[i].nlist[1];
    if (j<k){
      if (aux[j]>j)
	aux[j]=j;
      if (aux[k]>j)
	aux[k]=j;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    else{
      if (aux[j]>k)
	aux[j]=k;
      if (aux[k]>k)
	aux[k]=k;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    
    j=gt->gedges[i].nlist[2];
    k=gt->gedges[i].nlist[3];
    if (j<k){
      if (aux[j]>j)
	aux[j]=j;
      if (aux[k]>j)
	aux[k]=j;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    else{
      if (aux[j]>k)
	aux[j]=k;
      if (aux[k]>k)
	aux[k]=k;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }

  }
  
  v = new long [nn];
  for (i=0;i<nn;i++){
    v[i]=-1;
  }
  
  
  for (i=0;i<nn;i++){
    j=aux[i];
    if (j!=nn)
      v[j]++;
  }
  
  //  generation of interface node numbers
  //  number of interface nodes
  nin=0;
  for (i=0;i<nn;i++){
    if (v[i]>-1){
      v[i]=nin;
      nin++;
    }
  }
  
  for (i=0;i<nn;i++){
    j=aux[i];
    if (j!=nn)
      aux[i]=v[j];
    else
      aux[i]=-1;
  }

  
  //
  //  array ltg
  //
  if (ltg==NULL){
    ltg = new long* [ns];
    for (i=0;i<ns;i++){
      ltg[i] = new long [nnsd[i]];
    }
  }
  
  k=0;
  for (i=0;i<ns;i++){
    for (j=0;j<nnsd[i];j++){
      ltg[i][j]=aux[k];
      k++;
    }
  }
  
  
  fprintf (out,"\n\n\n funkce seqtop::create_ltg   pole ltg");
  for (i=0;i<ns;i++){
    fprintf (out,"\n");
    for (j=0;j<nnsd[i];j++){
      fprintf (out,"%ld  ",ltg[i][j]);
    }
  }
  fprintf (out,"\n");
  
  delete [] aux;
  delete [] v;
}
*/

/**
   function assembles array ltg1
   
   JK, 22.11.2007
*/
/*
void seqtop::ltg1array (long n,long *aux)
{
  long i;

  //  total number of nodes in the whole problem
  if (tnnp!=n)
    fprintf (stderr,"\n tnnp is not equal to n in function ltg1 (file %s, line %d)\n",__FILE__,__LINE__);
  tnnp = n;
  
  if (ltg1!=NULL)
    delete [] ltg1;
  ltg1 = new long [tnnp];

  for (i=0;i<tnnp;i++){
    ltg1[i] = aux[i];
  }
  
}
*/

/**
   function reads array ltg1
   
   @param n - total number of nodes in problem
   @param in - input file
   
   JK, 22.11.2007
*/
/*
void seqtop::readltg1 (long n,XFILE *in)
{
  long i;
  
  //  total number of nodes in the whole problem
  tnnp = n;
  
  if (ltg1!=NULL)
    delete [] ltg1;
  ltg1 = new long [tnnp];
  
  for (i=0;i<tnnp;i++){
    xfscanf (in,"%ld",ltg1+i);
  }
  
}
*/
