#ifndef SEQTOP_H
#define SEQTOP_H

#include <stdio.h>
#include "galias.h"
#include "xfile.h"

class gtopology;

/**
   class imitates the class partop which deals with topology in parallel computations
   it is sequential version of the class partop and it is developed
   for simpler manipulation and possible application of debugger
   
   
   notation:

   node multiplicity - number of subdomains which share the node

   internal node - node with multiplicity 1, (node inside a subdomain,
		   node not lying on inter-subdomain boundary)

   interface node - node with multiplicity at least 2, (node lying
                    on inter-subdomain boundary)

   coarse node - artificial node defined on master which collects appropriate
                 interface nodes

   
   global ordering - node ordering before mesh partitioning
   global node numbers - node numbers used in undecomposed mesh
                        (node numbers used before decomposition/partitioning)

   local ordering - ordering of subdomains (without respect to remaining subdomains)
   local node numbers - node numbers used in decomposed mesh
                        (node numbers used after decomposition/partitioning)

   coarse ordering - ordering of interface nodes only
   coarse node number - node numbers of interface nodes only, ordering
                        of all interface nodes in coarse problem

   group ordering - ordering of selected nodes only
   group node number - only selected nodes are ordered on the whole problem, it is similar to
                       the coarse ordering


   mesh description:
   md = 1 - all nodes have their global node number
   md = 2 - only interface nodes have coarse number, internal nodes are denoted by -1
   md = 3 - all nodes have their global node number, interface nodes have their global number multiplied by -1
   
   JK, 11.9.2007
*/
class seqtop
{
 public:
  seqtop(long nd,meshdescription meshd);
  ~seqtop ();

  void read_nnsd (XFILE *in);
  void read_ltg (XFILE *in);
  void read_loadbal_ltg (XFILE *in);
  void compute_multiplicity (FILE *out);
  void find_boundary_nodes (FILE *out);
  void rewrite_ltg ();
  void codnum_renumb (gtopology *top);

  //void create_ltg (gtopology *gt,FILE *out);
  //void ltg1array (long n,long *aux);
  //void readltg1 (long n,XFILE *in);

  ///  number of subdomains
  long ns;
  ///  total number of nodes in the whole problem
  long tnnp;
  ///  total number of interface (boundary) nodes
  long tnbn;

  ///  mesh description
  meshdescription md;
  
  ///  array containing number of nodes on subdomains
  ///  it contains nproc components
  ///  nnsd[i]=j - j nodes are defined on the i-th subdomain
  long *nnsd;

  ///  array containing list of numbers of boundary/interface nodes on subdomains
  ///  it contains ns components
  ///  nbnd[i]=j - the i-th domain contains j boundary/interface nodes
  long *nbnd;

  ///  array containing list of numbers of internal nodes on subdomains
  ///  it contains ns components
  ///  nind[i]=j - the i-th domain contains j internal nodes
  long *nind;

  ///  array containing numbers of boundary nodes
  ///  it contains ns, nbnd[i] components
  ///  lnbn[i][j]=k - the j-th boundary node on the i-th subdomain has local number k
  ///  array is allocated in function find_boundary_nodes
  long **lnbn;

  ///  array containing numbers of internal nodes
  ///  it contains ns, nind[i] components
  ///  lnin[i][j]=k - the j-th internal node on the i-th subdomain has local number k
  ///  array is allocated in function find_boundary_nodes
  long **lnin;

  ///  array containing list of global numbers of interface nodes
  ///  lgnbn[i][j]=k - the j-th interface node on the i-th subdomain has global number k
  long **lgnbn;

  ///  array containing number of subdomains which each boundary/interface node belongs to
  ///  it contains tnnp or tnbn components, it depends on the type of mesh description
  ///  multip[i]=j - the i-th node belongs to j subdomains
  long *multip;

  ///  array containing number of subdomains which share the nodes
  ///  it contains ns, nnsd[i] components
  ///  nodmultip[i][j]=k - the j-th node of the i-th subdomain belongs to k subdomains
  long **nodmultip;


  ///  local to global correspondence
  ///  ltg[i][j]=k - the j-th node on the i-th subdomain has coarse number k
  ///  k>-1 - number of coarse unknown
  ///  k=-1 - the node is internal node
  long **ltg;

};

#endif
