#include "siftop.h"
//#include "siftop_element_types.h"
//#include "cmlfile.h"
#include <stdlib.h>
#include <string.h>

/**
  This method allocates array node which size is given by osize

  @param osize - number of nodes of given element i.e. size of node array

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void selement::alloc(long osize)
{
  size=osize ;
  node=new long[size] ;
}


/**
  This constructor initializes data mebers with NULL or zero values

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
siftop :: siftop (void)
{
  E  = NULL;
  N  = NULL;
  Pn = NULL;
  nodes    = 0;
  elements = 0;
  
  //  parameters for the BOSS preconditioning
  nsd=0;
  meshtype=0;
  nnsd=NULL;
}


/**
  This destructor deallocates memory used by data mebers

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
siftop :: ~siftop(void)
{
  if (E)
    delete [] E;

  if (N)
    delete [] N;

  if (Pn)
    delete [] Pn;
  
  if (nnsd)
    delete [] nnsd;
}


/**
  This method allocates memory for array nodes which size is given by onodes

  @param onodes - number of nodes in given domain

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_number_of_nodes(long onodes)
{
  nodes = onodes;
  N = new snode[nodes];
}


/**
  This method allocates memory array Pn which size is given by vlaue of data member nodes

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: aloc_pnn(void)
{
  Pn = new long[nodes];
}


/**
  This method sets global node number of oid-th node of given domain to value given by oval

  @param oid  - node number. (Indexing starts from 0)
  @param oval - global node number

  @note Only nodes on boundary  of subdomain has assigned global node numbers. Rest of nodes hasn' t
  assigned global node number. This global node numbering is support for paralell computing.

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_paral_nn(long oid, long oval)
{
  if (oid >= nodes)
    return;
  Pn[oid] = oval;
}


/**
  This method sets number of elements of given domain to value given by oelements

  @param oelements - number of elements.

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_number_of_elements(long oelements)
{
  elements = oelements;
  E = new selement[elements];
}


/**
  This method sets coordinates of oid-th node

  @param oid - node number. (Indexing starts from 0)
  @param ox  - x coordinate value
  @param oy  - y coordinate value
  @param oz  - z coordinate value

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_node_coordinates(long oid, double &ox, double &oy, double &oz)
{
  if (oid >= nodes)
    return ;

  N[oid].x = ox;
  N[oid].y = oy;
  N[oid].z = oz;
}


/**
  This method sets property number of oid-th node

  @param oid - node number.  (Indexing starts from 0)
  @param oproperty - property value

  @note Property value of given node is used by preprocessor. Nodes with same property value
  have same properties. (For example ndofs, boundary conditions, . . .)

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_node_properties(long oid, long oproperty)
{
    if ( oid>nodes )
    {
        return ;
    }
    N[oid].property=oproperty ;
}


/**
  This method sets type number of oid-th element

  @param oid - element number. (Indexing starts from 0)
  @param otype - type number.

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_element_type (long oid, long otype)
{
  if (oid > elements)
    return ;
  E[oid].type = otype;
}


/**
  This method sets number of oid-th element edges

  @param oid - element number. (Indexing starts from 0)
  @param oned - number of edges.

  created by Tomas Koudelka 24.2.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_element_nedges(long oid, long oned)
{
  if (oid > elements)
    return ;
  E[oid].ned = oned;
  E[oid].propedg = new long[oned];
  memset(E[oid].propedg, 0, sizeof(*E[oid].propedg)*oned);
}

/**
  This method sets number of oid-th element surfaces

  @param oid - element number. (Indexing starts from 0)
  @param onsurf - number of surfaces.

  created by Tomas Koudelka 10.11.2005, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_element_nsurf(long oid, long onsurf)
{
  if (oid > elements)
    return ;
  E[oid].nsurf = onsurf;
  E[oid].propsurf = new long[onsurf];
  memset(E[oid].propsurf, 0, sizeof(*E[oid].propsurf)*onsurf);
}

/**
  This method sets node numbers of oid-th element and allocates memory for
  selement array node.

  @param oid - element number. (Indexing starts from 0)
  @param osize - number of element nodes.
  @param onode_index - pointer to array with node numbers of given element.

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_element_nodes(long oid, long osize, long *onode_index)
{
  if (oid > elements)
    return;
  E[oid].alloc(osize);
  memcpy(E[oid].node, onode_index, osize*sizeof(long));
}


/**
  This method sets property number of oid-th element

  @param oid - element number.  (Indexing starts from 0)
  @param oproperty - property value

  @note Property value of given element is used by preprocessor. Elements with same property value
  have same properties. (For example ndofs, material, cross-section, . . .)

  created by Ondra Hrstka 2000
  modified by Tomas Koudelka 24.1.2002, koudelka@cml.fsv.cvut.cz
*/
void siftop :: set_element_properties(long oid, long oproperty)
{
  if (oid > elements)
    return;
  E[oid].property = oproperty;
}

/**
  This method reads topology of given domain from file, which is in the JKTK format

  @param in   - pointer to structure with opened file of input data
  @param rgnn - indicator whether global node numbers array should be read (=1) or no (=0).
                This is support for paralell computing.

  @note JKTK mesh format is generated by programs gensif??? and details are in . . .

  created by Tomas Koudelka 11.2001, koudelka@cml.fsv.cvut.cz
*/
long siftop :: read(FILE *in, long rgnn, long redgn)
{
  long i, j, etype;
  if (fscanf(in, "%ld", &nodes) != 1)
  {
    fprintf(stderr, "\n\nError reading number of nodes\n");
    fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
    return(1);
  }
  N = new snode[nodes];
  memset(N, 0, sizeof(*N)*nodes);
  for (i = 0; i < nodes; i++)
  {
    if (fscanf(in, "%le %le %le %ld", &N[i].x, &N[i].y, &N[i].z, &N[i].property) != 4)
    {
      fprintf(stderr, "\n\nError reading node coordinates or property number\n");
      fprintf(stderr, " node number %ld\n", i+1);
      fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
      return(2);
    }
  }
  if (fscanf(in, "%ld %ld", &elements, &etype) != 2)
  {
    fprintf(stderr, "\n\nError reading number of elements or element type\n");
    fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
    return(3);
  }
  E = new selement[elements];
  memset(E, 0, sizeof(*E)*elements);
  for (i = 0; i < elements; i++)
  {
    if (fscanf(in, "%ld", &E[i].size) != 1)
    {
      fprintf(stderr, "\n\nError reading number of element nodes\n");
      fprintf(stderr, " element number %ld\n", i+1);
      fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
      return(4);
    }
    E[i].node = new long[E[i].size];
    E[i].type = etype;
    memset(E[i].node, 0, sizeof(*E[i].node)*E[i].size);
    for (j = 0; j < E[i].size; j++)
    {
      if (fscanf(in, "%ld", &E[i].node[j]) !=1)
      {
        fprintf(stderr, "\n\nError reading element node number %ld\n", j+1);
        fprintf(stderr, " element number %ld\n", i+1);
        fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
        return(5);
      }
    }
    if (fscanf(in, "%ld", &E[i].property) != 1)
    {
      fprintf(stderr, "\n\nError reading property number\n");
      fprintf(stderr, " element number %ld\n", i+1);
      fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
      return(6);
    }
    if (redgn)
    {
      if (fscanf(in, "%ld", &E[i].ned) != 1)
      {
        fprintf(stderr, "\n\nError reading number of element edges\n");
        fprintf(stderr, " element number %ld\n", i+1);
        fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
        return(7);
      }
      E[i].propedg = new long [E[i].ned];
      memset(E[i].propedg, 0, sizeof(*E[i].propedg)*E[i].ned);
      for (j = 0; j < E[i].ned; j++)
      {
        if (fscanf(in, "%ld", E[i].propedg+j) != 1)
        {
          fprintf(stderr, "\n\nError reading element edge number\n");
          fprintf(stderr, " element number %ld\n", i+1);
          fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
          return(8);
        }
      }
    }
  }
  
  
  long tnn=0;
  if (rgnn==0)
    return(0);
  if (rgnn==2){
    fscanf (in,"%ld %ld",&meshtype,&nsd);
    nnsd = new long [nsd];
    for (i=0;i<nsd;i++){
      fscanf (in,"%ld",nnsd+i);
      tnn+=nnsd[i];
    }
  }
  if (rgnn==1)
    tnn=nodes;
  
  //Pn = new long[nodes];
  //memset(Pn, 0, sizeof(*Pn)*nodes);
  Pn = new long[tnn];
  memset(Pn, 0, sizeof(*Pn)*tnn);
  
  for (i = 0; i < tnn; i++){
    if (fscanf(in, "%ld", &Pn[i]) != 1){
      fprintf(stderr, "\n\nError reading global number of nodes\n");
      fprintf(stderr, " node number %ld\n", i+1);
      fprintf(stderr, " function siftop::read() (%s, line %d)\n",__FILE__,__LINE__);
      return(7);
    }
  }
  return(0);
}

/*
void siftop::output_as_cml ( char *ofname )
{
    cmlfile f ;
    f.open_for_output( ofname ) ;

    f.set_labels( 3 ) ;
    f.set_label_string( 0,"ElementType" ) ;
    f.set_label_string( 1,"Nodes" ) ;
    f.set_label_string( 2,"Elements" ) ;

    f.set_sections( 2 ) ;
    f.set_section_string( 0,"Nodes" ) ;
    f.set_section_string( 1,"Elements" ) ;

    f.out( f.LABEL( "Nodes" ),nodes ) ;
    f.out( f.LABEL( "Elements" ),elements ) ;

    f.out_free_string( "# -- element types --" ) ;
    f.out_eoln() ;

    long written[AllElementTypes] ;
    long i,j ;
    for ( j=0 ; j<AllElementTypes ; j++ )
    {
        written[j]=0 ;
    }

    for ( i=0 ; i<elements ; i++ )
    {
        j=E[i].type ;
        if ( !written[j] )
        {
            f.out_label( f.LABEL( "ElementType" )) ;
            f.out_int( j ) ;
            f.out_string( siftop_element_type(j)) ;
            f.out_eoln() ;
            written[j]=1 ;
        }
    }

    f.out_free_string( "# -- nodes --" ) ;
    f.out_eoln() ;

    f.out_section( f.SECTION( "Nodes" )) ;

    for ( i=0 ; i<nodes ; i++ )
    {
        f.out_int( i ) ;
        f.out_double( N[i].x ) ;
        f.out_double( N[i].y ) ;
        f.out_double( N[i].z ) ;
        f.out_int( N[i].property ) ;
        f.out_eoln() ;
    }

    f.out_section_end() ;

    f.out_free_string( "# -- elements --" ) ;
    f.out_eoln() ;

    f.out_section( f.SECTION( "Elements" )) ;

    for ( i=0 ; i<elements ; i++ )
    {
        f.out_int( i ) ;
        f.out_int( E[i].type ) ;
        f.out_int( E[i].property ) ;
        for ( j=0 ; j<E[i].size ; j++ )
        {
            f.out_int( E[i].node[j] ) ;
        }
        f.out_eoln() ;
    }

    f.out_section_end() ;

    f.out_free_string( "# -- eof --" ) ;
    f.out_eoln() ;
}
*/

void siftop::exporttop (gtopology *top)
{
  long i, j;
  top->alloc_nodes(nodes);
  top->alloc_elements(elements);
  for (i = 0; i < nodes; i++)
  {
    top->gnodes[i].x = N[i].x;
    top->gnodes[i].y = N[i].y;
    top->gnodes[i].z = N[i].z;
  }
  for (i = 0; i < elements; i++)
  {
    top->gelements[i].nne = E[i].size;
    top->gelements[i].nodes = new long[E[i].size];
    for (j = 0; j < E[i].size; j++)
      top->gelements[i].nodes[j] = E[i].node[j];
  }
}

/**
   function prints mesh data in SSMF
   
   JK, 7.3.2007 
*/
void siftop::print (FILE *out)
{
  long i, j;
  
  fprintf (out,"%ld\n",nodes);

  for (i = 0; i < nodes; i++)
    fprintf (out,"%15.10le %15.10le %15.10le %ld\n",N[i].x,N[i].y,N[i].z,N[i].property);
  
  fprintf (out,"%ld %ld\n",elements,E[0].type);
  
  for (i = 0; i < elements; i++){
    fprintf (out,"%ld ",E[i].size);
    for (j = 0; j < E[i].size; j++)
      fprintf (out,"%ld ",E[i].node[j]);
    fprintf (out,"%ld\n",E[i].property);
  }
  
}
