#if !defined ( __SIFTOP__ )
#define        __SIFTOP__

#include <stdio.h>
#include "gtopology.h"

struct snode
/**
  This structure groups information about node like coordinates and property number
*/
{
    double x; ///< x coordinate of node
    double y; ///< y coordinate of node
    double z; ///< z coordinate of node
    long property ; ///< property number of node
};

struct selement
/**
  This structure groups information about element like node numbers, property number and type of element
*/
{
    long *node; ///< pointer to array with element node numbers
    long size;  ///< size of array node
    long type;  ///< type number of element
    long property; ///< property number of element
    long ned;      ///< number of edges
    long nsurf;    ///< number of surfaces
    long *propedg; ///< property numbers of each edge
    long *propsurf;///< property numbers of surfaces
    void alloc(long osize); ///< allocation of memory for nodes
};


class siftop
/**
  This class stores all information about topology of given problem but only for auxiliary
  purposes.
*/
{
    public:
     siftop(void); ///< constructor
     ~siftop(void ) ; ///< destructor

     /// method sets number of nodes
     void set_number_of_nodes(long onodes);

     /// method sets number of elements
     void set_number_of_elements(long oelements);

     /// method allocates array with global node numbers
     void aloc_pnn(void);

     /// method sets number of boundary nodes
     void set_paral_nn(long oid, long oval);

     /// method sets coordinates of given node
     void set_node_coordinates(long oid, double &ox, double &oy, double &oz);

     /// method sets property number of given node
     void set_node_properties(long oid, long oproperty);

     /// method sets element type
     void set_element_type(long oid, long otype);

     /// method sets number of element edges
     void set_element_nedges(long oid, long oned);

     /// method sets number of element surfaces
     void set_element_nsurf(long oid, long onsurf);

     /// method sets node numbers of given element
     void set_element_nodes(long oid, long osize, long *onode_index);

     /// method sets property number of given element
     void set_element_properties(long oid, long oproperty);

     /// method reads topology in JKTK format from file
     long read(FILE *in, long rgnn, long redgn);

//     void output_as_cml(char *ofname);
     
     void exporttop (gtopology *top);
     
     void print (FILE *out);


     selement *E ;  ///< pointer to array with elements
     snode *N ;     ///< pointer to array with nodes
     long nodes;    ///< number of nodes
     long elements; ///< number of elements
     long *Pn;      ///< Pointer to array with global node numbers. This is support for paralell computing
       
     //  parameters for BOSS preconditioners
     ///  type of mesh description
     long meshtype;
     ///  number of subdomains
     long nsd;
     /// number of nodes on subdomains
     long *nnsd;
} ;

#endif
