#ifndef TABLEFUNCT_H
#define TABLEFUNCT_H

#include <stdio.h>
#include "iotools.h"


enum interpoltype{piecewiselin=1,dirac=2,lagrange=3,piecewiselin2=4,dirac2=5};///<supported types of interpolation
const enumstr interpoltypestr[5] = {{"linear",1}, {"dirac",2}, {"lagrange",3}, {"linear2",4}, {"dirac2", 5}};
const kwdset interpoltype_kwdset(sizeof(interpoltypestr)/sizeof(*interpoltypestr), interpoltypestr);

/**
   This file declares the class of tablefunct, which implements various types of interpolation.
   The values for interpolation are given by the table.
*/
class tablefunct
{
 public:
  tablefunct();
  ~tablefunct();
  double getval(double temp);
  double getval2(double temp, double &k);
  double getval3(double temp, double &k);
  void   read  (XFILE *in);
  void print(FILE *out);
  void read_prop(FILE *in);

  void copy (tablefunct &tf);

  ///  type of interpolation
  interpoltype itype;
  ///  number of function values
  long asize;
  ///  array of independent variable (usually denoted x)
  double *x;
  ///  array of function values (dependent variable, usually denoted y)
  double *y;
  
  double lininterpol  (double temp, double &k);
  
 private:
  double diracinterpol(double temp);
  double diracinterpol2(double temp, double &k);
  double lininterpol  (double temp);
  double lininterpol2  (double temp, double &k);
  double lininterpol3  (double temp, double &k);
  double laginterpol  (double temp);

};
#endif
