#include "timecontr.h"

timecontr::timecontr (void)
{
  //  actual time
  time=0.0;
  //  time of start of analysis
  start_time=0.0;
  //  time of end of analysis
  end_time=0.0;

  //  forward time step
  fdt=0.0;
  //  backup of forward time step
  bfdt=0.0;
  //  backward time step
  bdt=0.0;
  
  //  number of important times
  nit=0;
  //  actual position in important times
  apit=0;
  //  array containing important times
  imptime=NULL;
  //  indicator of important times
  iiit=0;
  
  //  type of time controller
  tct=fixed;
  //  minimum time step
  dtmin=0.0;
  //  maximum time step
  dtmax=0.0;
}

timecontr::~timecontr (void)
{
  delete [] imptime;
}

/**
   function reads necessary information about time measurement
   
   @param in - input stream
   
   6.4.2003, JK
*/
void timecontr::read (XFILE *in)
{
  long i;
  
  //  type of time controller
  xfscanf (in,"%k%m","time_contr_type",&timecontrtype_kwdset,(int*)&tct);
  
  //  starting and ending times
  xfscanf (in,"%k%lf%k%lf","start_time",&start_time,"end_time",&end_time);
  
  //  number of important times
  xfscanf (in,"%k%ld","num_imp_times", &nit);
  
  //  important times
  // pokus JM 29.5.2008
  if (nit == -1){
    double krok;
    xfscanf (in,"%lf",&krok);
    nit = long((end_time - start_time)/krok);
    if (nit < 2)
      nit = 2;
    
    imptime=new double [nit];
    imptime[0] = start_time + krok;
    
    for (i=1; i<nit-1; i++)
      imptime[i] = imptime[i-1] + krok;
    imptime[nit-1] = end_time;
  }
  else{
    if (nit==0){
      imptime = new double [1];
      imptime[0]=end_time;
    }
    else{
      imptime=new double [nit];
    }
    for (i=0;i<nit;i++){
      xfscanf (in,"%lf",imptime+i);
    }
  }

  //  function prescribing time increments/steps
  timefun.read (in);

  switch (tct){
  case fixed:{
    break;
  }
  case adaptive:{
    xfscanf (in,"%k%lf %k%lf","dtmin",&dtmin,"dtmax",&dtmax);
    break;
  }
  case adaptivemin:{
    xfscanf (in,"%k%lf","dtmin",&dtmin);
    break;
  }
  case adaptivemax:{
    xfscanf (in,"%k%lf","dtmax",&dtmax);
    break;
  }
  default:{
    print_err("unknown type of time controller is required",__FILE__,__LINE__,__func__);
  }
  }
  
  
  //  initialization of actual time
  time = start_time;
  //  initialization of the first forward time increment
  fdt = initialtimeincr ();
  //  backup of forward time increment
  bfdt = fdt;
}

/**
   function prints necessary information about time measurement
   
   @param out - output stream
   
   6.4.2003, JK
*/
void timecontr::print (FILE *out)
{
  long i;
  
  //  starting and ending times
  fprintf (out," %e %e\n",start_time,end_time);
  
  //  number of important times
  fprintf (out," %ld \n",nit);
  
  //  important times
  for (i=0;i<nit;i++){
    fprintf (out," %f \n",imptime[i]);
  }
  fprintf (out,"\n");
  
  //  function prescribing time increments/steps
  timefun.print (out);
  
}


/**
   function returns new time with respect to important times
   
   6.4.2003, JK
*/
double timecontr::newtime ()
{
  double newtime;
  
  //  trial time increment
  fdt=timefun.getval(time);
  
  //  trial new time
  newtime=time+fdt;
  
  //  indicator of important time
  iiit=0;
  if (apit<nit){
    if (imptime[apit]<=newtime){
      newtime=imptime[apit];
      iiit=1;
      apit++;
    }
    
    //  actual time step
    fdt=newtime-time;
    //  new time
    time=newtime;
  }
  else{
    //  new time
    time=newtime;
  }
  
  //  new backward time step
  bdt=fdt;
  return time;
}


/**
   function returns new time with respect to important times
   and time step enforced by an analysis
   
   @param dt - required time step from an analysis
   
   30.5.2008, JK
*/
double timecontr::newtime (double &dt)
{
  double newtime;
  
  if (iiit==1){
    //  previous time has been important time
    //  theoretically, the time step may be extremely small
    //  therefore, the time step before important time is used
    dt=bfdt;
  }

  //  trial time step
  fdt=timefun.getval(time);
  
  //TKr??!!
  //if (dt<fdt){
  //  time step has to be reduced with respect to precsribed dt
  //fdt=dt;
  //}
  
  //  backup of forward time step
  bfdt=fdt;

  //  trial new time
  newtime=time+fdt;
  
  //  indicator of important time
  iiit=0;
  if (apit<nit){
    if (imptime[apit]<=newtime){
      //fprintf (stdout,"\n\n jsme v important time %ld\n\n",apit);
      newtime=imptime[apit];
      iiit=1;
      apit++;
    }
    
    //  actual time step
    fdt=newtime-time;
    //  new time
    time=newtime;
    
  }
  else{
    //  new time
    time=newtime;
  }
  
  //  new backward time step
  bdt=fdt;
  //dt=fdt;TKr!!??
  return time;
}


/**
   function shifts time to the previous time instant
   
   29.5.2008, JK
*/
void timecontr::oldtime ()
{
  if (apit>0){
    if (imptime[apit-1]==time){
      apit--;
    }
  }
  
  //  back shift in time to the previous value
  time-=bdt;
  
  if (imptime[apit]==time)
    iiit=1;
  else
    iiit=0;
  
}

/**
   function returns starting time
   
   6.4.2003, JK
*/
double timecontr::starttime ()
{
  return start_time;
}

/**
   function returns ending time
   
   6.4.2003, JK
*/
double timecontr::endtime ()
{
  return end_time;
}

/**
   function returns actual time
   
   6.4.2003, JK
*/
double timecontr::actualtime ()
{
  return time;
}

/**
   function returns actual forward time increment
   
   6.4.2003, JK
*/
double timecontr::actualforwtimeincr ()
{
  return fdt;
}

/**
   function returns actual backward time increment
   
   6.4.2003, JK
*/
double timecontr::actualbacktimeincr ()
{
  return bdt;
}

/**
   function returns initial time increment
   
   6.4.2003, JK
*/
double timecontr::initialtimeincr ()
{
  double dt;

  dt = timefun.getval(start_time);
  
  if (nit==0){
    return dt;
  }
  else{
    if (start_time+dt<imptime[0]){
      return dt;
    }
    else{
      dt = imptime[0]-start_time;
      return dt;
    }
  }
}

/**
   function takes values from another time controller
   
   @param timecontr - pointer to another time controller
   
   JK, 20.9.2004
*/
void timecontr::take_values (timecontr &tc)
{
  time = tc.time;
  bdt = tc.bdt;
  fdt = tc.fdt;
  start_time = tc.start_time;
  end_time = tc.end_time;
}


/**
   function transforms time in seconds to time in days
   
   JK, 20.9.2004
*/
void timecontr::seconds_days ()
{
  time = time/86400.0;
  bdt = bdt/86400.0;
  fdt = fdt/86400.0;
  start_time = start_time/86400.0;
  end_time = end_time/86400.0;
}

/**
   function returns indicator of important time
   
   iiit=0 - it is not important time
   iiit=1 - it is important time
*/
long timecontr::isitimptime ()
{
  return iiit;
}

/**
   function sets up input data from other object of the class timecontr
   
   @param tc - another time controller
   
   JK
*/
void timecontr::initiate (timecontr &tc)
{
  long i;
  
  start_time = tc.start_time;
  end_time = tc.end_time;
  
  timefun.initiate (tc.timefun);

  nit = tc.nit;
  imptime = new double [nit];
  for (i=0;i<nit;i++){
    imptime[i]=tc.imptime[i];
  }
  
  time = start_time;

  //  initialization of the first forward time increment
  fdt = initialtimeincr ();
}

/**
   function reduces time step with respect to external conditions
   
   this function is used in nonlinear dynamics in cases of
   nonequilibriated stages
   
   @param rf - reduction factor
   
   in the case of reduction three times, rf = 3;
   
   10.9.2007, JK
*/
/*
void timecontr::reduce_step (double rf)
{
  //  time shift back
  time -= fdt;

  //  reduction of time step
  if (rf<1.0e-3){
    fprintf (stderr,"\n nonpositive reduction factor in function reduce_step (file %s, line %d).\n",__FILE__,__LINE__);
  }
  fdt /= rf;
  
  //  new time
  time += fdt;
}
*/
